﻿;;;-*- Mode: Lisp; Package: "CSCWD-MEMBERS" -*-
;;;===============================================================================
;;;14/09/16
;;;                AGENT MEMBERS (file: CSCWD-MEMBERS.lisp)
;;;                Copyright Barthès@UTC 2013
;;;
;;;===============================================================================
;;; This agent is part of the ONTOCODESSIGN application and contains the list of
;;; people who are authorized to access the information 
;;; It is called by the web agent to authorize logins
#|

2014
 0916 creation
|#

(defpackage :CSCWD-MEMBERS (:use :moss :omas :cl)(:nicknames :cm))
(in-package :CSCWD-MEMBERS)

(omas::defagent :CSCWD-MEMBERS :persistency nil)

;;;=============================================================================
;;;================================== Globals ==================================
;;;=============================================================================


;;;=============================================================================
;;;============================= Service Functions =============================
;;;=============================================================================

(defun make-person-code (id)
  "builds a code from the person internal sequence number, the first-name and ~
   name." 
   (format nil "~A-~{~A~^-~}/~{~A~^-~}"
     (moss::%get-internal-instance-number id)
     (send id '=get "first name")
     (send id '=get "name")))


#|
(make-person-code '$e-member.24)
"24-Susan/Finger"
|#
;;;=============================================================================
;;;================================== Skills ===================================
;;;=============================================================================

;;;======================================================================= skill
;;;                               AUTHORIZED?
;;;=============================================================================

(defskill :authorized? :CSCWD-MEMBERS
  :static-fcn static-authorized?
  )

(defun static-authorized? (agent message args)
  "verifies the login name. Assumes that names are different."
  (declare (ignore message))
  ;(format t "~%; static-authorized? /args: ~S " args)
  (let* ((data (cdr (assoc :data args)))
         ;(language (cdr (assoc :language args))) ; not necessary
         (*language* :en)
         (login (getstr "login" data))
         (password (getstr "password" data))
         (pattern (cdr (assoc :pattern args)))
         person-id passwd result)
    
    ;; compares name and password to the ones we have here
    (setq person-id (car (access `("member" ("login" :is ,login)))))
    
    (when person-id
      (setq passwd (car (send person-id '=get "password")))
      (setq result
            (when (equal+ password passwd)
              (if pattern (cdr (fill-pattern person-id pattern))
                t))))
    
    (static-exit agent result)))

#|
(static-authorized? SA_CSCWD-MEMBERS nil '((:data ("login" . "barthes")
                                            ("password" . "barthes"))))
T
NIL
(static-authorized? SA_CSCWD-MEMBERS nil 
                    '((:data ("login" . "Barthes")
                             ("password" . "barthes"))
                      (:pattern "personne" ("nom")("prénom")("email"))))
(("nom" "Barthès") ("prénom" "Jean-Paul") ("email" "barthes@utc.fr"))
NIL

(static-authorized? SA_CSCWD-MEMBERS nil 
                    '((:data ("username" . "Barthès")
                             ("password" . "barthès"))
                      (:pattern "person" ("name")("first name")("email"))))
NIL
NIL

(static-authorized? SA_CSCWD-MEMBERS nil '((:data ("login" . "Abel")
                                            ("password" . "leba"))))
NIL
NIL

(static-authorized? SA_CSCWD-MEMBERS nil '((:data ("login" . "Abel")
                                            ("password" . "abel"))))
T
NIL

(static-authorized? SA_CSCWD-MEMBERS nil 
                    '((:data ("login" . "Abel")
                             ("password" . "abel"))
                      (:pattern "personne" ("nom")("prénom")("email"))))
(("nom" "Abel") ("prénom" "Marie-Hélène") ("email" "marie-helene.abel@utc.fr"))
NIL
|#
;;;======================================================================= skill
;;;                          CHECK PASSWORD
;;;=============================================================================

(defskill :check-password :CSCWD-MEMBERS
  :static-fcn static-check-password)

(defun static-check-password (agent message args)
  "checks the validity of the password"
  (declare (ignore message))
  (let* ((data (cdr (assoc :data args)))
         (passwd (getstr "password" data))
         (name (getstr "login" data))
         (*language* :en)
         person-id result recorded-password)
    
    (setq person-id (car (access `("member" ("login" :is ,name)))))
    
    (when person-id
      (setq recorded-password (car (send person-id '=get "password")))
      (setq result (equal+ passwd recorded-password))
      )
    
    (static-exit agent (or result :failure))))

#|
(static-check-password sa_cscwd-members nil '((:data ("login" . "Abel")
                                             ("password" . "leba"))))
:FAILURE
NIL
(static-check-password sa_cscwd-members nil '((:data ("login" . "Abel")
                                               ("password" . "abel"))))
T
NIL
|#
;;;======================================================================= skill
;;;                          GET ALL MEMBERS
;;;=============================================================================

(defskill :get-all-members :CSCWD-MEMBERS
  :static-fcn static-get-all-members
  )
  
(defun static-get-all-members (agent message &optional args)
  "return a list of all members"
  (declare (ignore message))
  (let ((pattern (cdr (assoc :pattern args)))
        answer id-list)
    (setq id-list (access '("person")))
    ;; shape the result
    (setq answer (if pattern
                     (fill-pattern id-list pattern)
                   (broadcast id-list '=make-string)))
    (static-exit agent (or answer :failure))))
	
#|
(static-get-all-members SA_MEMBERS nil nil)
(("Abel: Marie-Hélène") ("Bonnifait: Philippe") ("Bouabdallah: Abdelmadjid") ("Charara: Ali")
 ("Barthès: Jean-Paul") ("Fontaine: Dominique") ("Fremont: Vincent") ("Lenne: Dominique")
 ("Lourdeaux: Domitile") ("Morizet-Mahoudeaux; Morizet: Pierre") ...)
 
(static-get-all-members sa_MEMBERS nil 
              '((:pattern "personne" ("nom")("prénom") ("organisation"))))
(("personne" ("nom" "Abel") ("prénom" "Marie-Hélène") ("organisation" "UTC"))
 ("personne" ("nom" "Bonnifait") ("prénom" "Philippe") ("organisation" "UTC"))
 ("personne" ("nom" "Bouabdallah") ("prénom" "Abdelmadjid") ("organisation" "UTC"))
 ("personne" ("nom" "Charara") ("prénom" "Ali") ("organisation" "UTC"))
 ("personne" ("nom" "Barthès") ("prénom" "Jean-Paul") ("organisation" "UTC"))
 ("personne" ("nom" "Fontaine") ("prénom" "Dominique") ("organisation" "UTC"))
 ("personne" ("nom" "Fremont") ("prénom" "Vincent") ("organisation" "UTC"))
 ("personne" ("nom" "Lenne") ("prénom" "Dominique") ("organisation" "UTC"))
 ...)
|#

;;;======================================================================= skill
;;;                            GET EMAIL
;;;=============================================================================
;;; called from dialog to obtain object description

(defskill :get-email :CSCWD-MEMBERS
  :static-fcn static-get-email)

(defun static-get-email (agent message args)
  "returns email if person exists"
  (declare (ignore message))
  (let* ((*language* (or (getvl args :language) *language*))
         (data (getvl args :data))
         (name (get-field "name" data))
         (first-name (get-field "first name" data))
         id)
    
    (setq id 
          (car
           (access `("person" 
                     ("name" :is ,@name)(first-name :is ,@first-name)))))
    (static-exit agent (when id (car (send id '=get "email"))))))

#|
(static-get-email sa_members :dummy '((:language . :EN)
                                      (:data ("name" "Barthès")
                                             ("first name" "Jean-Paul"))))
"barthes@utc.fr"
|#
;;;======================================================================= skill
;;;                          GET MEMBER CODE
;;;=============================================================================
;;; from the name and first name retrieves the member code (author-id)

(defskill :get-member-code :CSCWD-MEMBERS
  :static-fcn static-get-member-code
  )
  
(defun static-get-member-code (agent message args)
  "return a list of all members"
  (declare (ignore message))
  (let* ((data (cdr (assoc :data args)))
         (name (car (getstr "nom" data)))
         (first-name (car (getstr "prénom" data)))
         id result)
    ;; try to get member
    (setq id (car (access 
              `("member" ("nom" :is ,name)
                ,@(if first-name `(("prénom" :is ,first-name)))))))
    
    (if id (setq result (make-person-code id)))
    (if result
        (static-exit agent result)
      (static-exit agent :failure :reason "Je n'ai personne de ce nom-là."))))
		
#|
(static-get-member-code SA_MEMBERS nil '((:data ("nom" "Abel")
                                                ("prénom" "Marie-Hélène"))))
"1-Marie-Hélène/Abel"
NIL
|#
;;;X====================================================================== skill
;;;                          GET MEMBER DESCRIPTION
;;;=============================================================================
;;; returns a description list of a member according to the language 

(defskill :get-member-description :CSCWD-MEMBERS
   :static-fcn static-get-member-description)

(defun static-get-member-description (agent message args)
  "gets a description list of a member according to the language. If data is not ~
   empty, then considers that it is the name of the personal assistant, and ~
   returns the property tagged along.
Arguments:
   agent: current agent
   message: unused
   args: e.g. ((:data :master \"Ramos\" :assistant \"Albert\")(:language . :fr))
Return:
   a list containing strings 
      (\"person\" (\"name\" \"Ramos\")
                  (\"first name\" \"Milton\")
                  (\"personal assistant\" \"Ze\"))"
  (declare (ignore message))
  (format t "~%;--- args: ~S" args)
  (let* ((data (cdr (assoc :data args)))
         (master (getf data :master))
         (assistant (getf data :assistant))
         (language (cdr (assoc :language args)))
         master-id description)
    (format t "~%;--- master login: ~S" master)
    ;; get the person id
    (setq master-id (car (access `("member" ("login" :is ,master)))))
    (format t "~%;--- master-id: ~S" master-id)
    
    (when master-id
      ;; get info from the ontology
      (setq description 
            (send master-id '=make-local-description language assistant)))
    
    (format t "~%<<< get-member-description")
    (format t "~%;--- returning: ~S" description) 
    (static-exit agent description)))
		  
#|
(static-get-member-description
 sa_cscwd-members nil '((:data :master "Abel" :assistant "Julie")(:language . :FR)))
("personne" ("nom" "Abel") ("prénom" "Marie-Hélène") ("assistant personnel" "Julie"))
(static-get-member-description
 sa_cscwd-members nil '((:data :master "Abel" :assistant "Julie")(:language . :EN)))
("person" ("name" "Abel") ("first name" "Marie-Hélène") ("personal assistant" "Julie"))
|#
;;;======================================================================= skill
;;;                          GET MEMBER EMAIL
;;;=============================================================================
;;; not yet implemented. Assumes that members accept to give their e-mail.
;;; Emails should be recorded and protected. It should be possible to send an 
;;; email but not see the actual member's email.
;;; Therefore, this skill is for internal use only.


;;;======================================================================= skill
;;;                          GET MEMBERS INFO
;;;=============================================================================
;;; send the list of members and their associated code

(defskill :get-members-info :CSCWD-MEMBERS
  :static-fcn static-get-members-info
  )
  
(defun static-get-members-info (agent message &optional args)
  "return a list of all members"
  (declare (ignore message))
  (let ((member-list (access '("Member")))
         result)
   ;; make a list of name / identifier
   (setq result
     (loop for item in member-list
         collect `(,@(send item '=make-string) ,(make-person-code item) )))
  (static-exit agent (or result :failure :reason "Can't access memebers."))))
		
#|
(static-get-members-info SA_MEMBERS nil)
(("Abel: Marie-Hélène" "1-Marie-Hélène/Abel") ("Bonnifait: Philippe" "2-Philippe/Bonnifait")
 ("Bouabdallah: Abdelmadjid" "3-Abdelmadjid/Bouabdallah") ("Charara: Ali" "4-Ali/Charara")
 ("Barthès: Jean-Paul" "5-Jean-Paul/Barthès") ("Fontaine: Dominique" "6-Dominique/Fontaine")
 ("Fremont: Vincent" "7-Vincent/Fremont") ("Lenne: Dominique" "8-Dominique/Lenne")
 ("Lourdeaux: Domitile" "9-Domitile/Lourdeaux")...)
|#	 

;;;======================================================================= skill
;;;                          SET PASSWORD
;;;=============================================================================

(defskill :set-password :CSCWD-MEMBERS
  :static-fcn static-set-password)

(defun static-set-password (agent message args)
  "sets the password to a new value"
  (declare (ignore message))
  (let* ((data (cdr (assoc :data args)))
         (passwd (getstr "password" data))
         (login (getstr "login" data))
         (*language* :en)
         person-id result)
    (setq person-id (car (access `("member" ("login" :is ,login)))))
    
    (when person-id
      (send person-id '=set "password" passwd)
      (setq result :done)
      )
    
    (static-exit agent (or result :failure))))

#|
(static-set-password sa_cscwd-members nil '((:data ("login" . "Abel")
                                             ("password" . "leba"))))
:DONE
NIL
|# 
;;;======================================================================= skill
;;;                              UNAUTHORIZE
;;;=============================================================================

;;; ================================= environment ==============================


:EOF