﻿;;;-*- Mode: Lisp; Package: "XXXXX" -*-
;;;==============================================================================
;;;14/07/22
;;;             AGENT XXXXX: TASKS (file XXXXX-tasks-EN.lisp)
;;;             Copyright Barthes@UTC, 2014
;;;
;;;==============================================================================

#|
This file contains the tasks shared by the English speaking PAs of the NEWS appli-
cation.

Tasks are currently:
ADD CATEGORY (NEWS)
ADD INFO (NEWS)
ADD KEYWORD (NEWS)
DELETE KEYWORD (NEWS)
DELETE NEWS ITEM$ (NEWS)
EXPLAIN
GETTING ALL CATEGORIES (NEWS) 
GETTING CATEGORIES (NEWS)
GETTING KEYWORDS (NEWS)
GETTING LAST NEWS (NEWS)
GETTING NEWS (NEWS)
HELP
PUBLISH-PRIVATE-NOTE
SET FONTS
SUBSCRIBE (NEWS)
UNSUBSCRIBE (NEWS)
WHAT-IS
WHO-IS

2014
 0303 cloned from the STEVENS-TASKS file
 0722 upgraded and cleaned
|#

;(in-package :XXXXX) 

(format t "~&;*** loading English tasks, package is: ~S" *package*)

;;;==============================================================================
;;;   
;;;                               TASK MODEL
;;;
;;;==============================================================================

;;; As soon as the PA finds a task to perform it puts its model into the 
;;; conversation object. 
;;; The CONVERSATION object will contain the info corresponding to the task
;;; arguments, filled during the specific conversation attached to the task.

(defconcept (:en "task index")
  (:doc :en "A TASK INDEX is a property of a task. The index attribute specifies ~
             the term being used, the weight attribute gives its importance for the ~
             task.")
  (:att (:en "index"))
  (:att (:en "weight"))
  )

(definstmethod =summary TASK-INDEX ()
  (list (format nil "~{~A ~}: ~{~S~^ ~}" (has-index)(has-weight))))

(defconcept (:en "task")
  (:is-a moss::$QTSKE) ; used for linking inheritage ?
  (:doc :en "A TASK is a model of task to be performed by the assistant.")
  (:att (:en "task name") (:unique)(:entry))
  (:att (:en "dialog"))
  (:att (:en "where to ask")) ; e.g. :ADDRESS
  (:att (:en "fail question"))
  (:rel (:en "index pattern")(:to "task index"))
  (:att (:en "performative") (:id $PRFT)(:one-of :request :assert :command :answer))
  (:att (:en "message action")) ; e.g. :get-address
  )

(definstmethod =summary TASK ()
  (HAS-TASK-NAME))
	 
;;;==============================================================================
;;;   
;;;                               TASK LIBRARY
;;;
;;;============================================================================== 

;;;================================================================= ADD CATEGORY
  
(deftask "NEWS: add category"
  :doc "Task for creating a new category."
  :performative :command
  :dialog _add-category-conversation
  :indexes ("create" .3 "new" .3 "add" .3 "category" .3)  
  )
  
;;;================================================================== ADD KEYWORD
  
(deftask "NEWS: add keyword"
  :doc "Task for creating a new keyword."
  :performative :command
  :dialog _add-index-conversation
  :indexes ("create" .3 "new" .3 "add" .3 "keyword" .3 "index" .3 "delete" -1)  
  )
  
;;;================================================================ ADD NEWS ITEM
  
(deftask "NEWS: add news item"
  :doc "Task for creating a new news item."
  :performative :command
  :dialog _add-info-conversation
  :indexes ( "new" .3 "add" .3 "news" .3 "item" .3 "information" .3
            "info" .3 "delete" -1)  
  )

;;;========================================================================== ASK

;;;(deftask "ask agent"
;;;  :doc  "Tâche (ordre) pour demander une information à un autre agent"
;;;  :performative :command
;;;  :dialog _ask-conversation
;;;  :indexes ("demander à" .8 "demander" .7 "demande" .7
;;;            "demandez" .7 "message pour" .7 "message" .4)
;;;  )

;;;================================================================= CREATE INFO
;;; this is a test for invoking a web page

(deftask "NEWS: create info"
  :doc "Task for invoking the news editor"
  :performative :command
  :dialog _create-info-conversation
  :indexes ("create" .6 "info" .3)  
  )

;;;============================================================== DELETE KEYWORD

(deftask "NEWS: delete keyword"
  :doc "Task for locally deleting a keyword."
  :performative :command
  :dialog _delete-keyword-conversation
  :indexes ("erase" .3 "kill" .3 "remove" .3 "delete" .3 "keyword" .3 "index" .3
             "add" -1)
  )
  
;;;============================================================ DELETE NEWS ITEM
;;; Deleting a news item is no longuer allowed

(deftask "NEWS: delete news item"
  :doc "Task for removing a provate info"
  :performative :command
  :dialog _delete-news-item-conversation
  :indexes ("delete" .3 "erase" .3 "kill" .3 "info" .3 "news item" .3 "item" .2
             "add" -1)
  )
  
;;;===================================================================== EXPLAIN 

;;;(deftask "define concept"
;;;  :doc "Tâche (ordre) donnant la définition d'un concept de l'ontologie."
;;;  :performative :command
;;;  :dialog _explain-conversation
;;;  :indexes ("définir" .6)
;;;  )

;;;============================================================= GETTING AUTHORS 

(deftask "NEWS: get authors"
  :doc "Task (question) for getting the authors of news items."
  :dialog _get-all-authors-conversation
  :indexes ("authors" .3 "persons" .3 "all" .2 "list" .2 )
  )
  
(deftask "NEWS: get authors"
  :doc "Task (command) for getting the authors of news items."
  :performative :command
  :dialog _get-all-authors-conversation
  :indexes ("authors" .3 "persons" .3 "all" .2 "list" .2 )
  )
  
;;;=========================================================== GETTING CATEGORIES

(deftask "NEWS: get categories"
    :doc "Task (question) for getting the list of categories."
  :dialog _get-all-categories-conversation
  :indexes ("show" .3 "category" .5 "categories" .5 "topics" .5 "all" .2)
  )

(deftask "NEWS: get categories"
    :doc "Task (command) for getting the list of categories."
  :performative :command
  :dialog _get-all-categories-conversation
  :indexes ("category" .3 "categories" .3 "topics" .3 "show" .25 "all" .2
             "display" .25 "print" .25 "list" .25 "delete" -1)
  )

;;;================================================================ GETTING EMAIL

;;;(deftask "get email"
;;;  :doc "Tâche permettant d'obtenir l'adresses email de quelqu'un"
;;;  :dialog _get-email-conversation
;;;  :indexes ("adresse" .2 "email" .5 "email de" .7 "mel" .7
;;;            "adresse électronique" .7)
;;;  )

;;;============================================================ GETTING KEYWORDS

(deftask "NEWS: get news keywords"
  :doc "Task (question) for getting the list of keywords to index news items."
  :dialog _get-all-indexes-conversation
  :indexes ("keywords" .4)
  )

(deftask "NEWS: get all keywords"
    :doc "Task (command) for getting the list of keywords."
  :performative :command
  :dialog _get-all-indexes-conversation
  :indexes ("keyword" .3 "keywords" .3 "index" .3 "show" .25 "all" .2
             "display" .25 "print" .25 "list" .25 "delete" -1)
  )

;;;===================================================== GETTING LAST LOCAL NEWS

(deftask "NEWS: get last local news"
  :doc "Task (question)for getting the last local news."
  :dialog _get-last-local-news-conversation
  :indexes ("view" .3 "news" .3 "new" .3 "last" .25 "local" .2 "posted" .3 
             "gave" .3 "show" .3 "personal" .2 "private" .2 
			 "notes" .3 "delete" -1 "add" -1)
  )

(deftask "NEWS: get last local news"
  :doc "Task (command) for getting the last local news."
  :dialog _get-last-local-news-conversation
  :performative :command
  :indexes ("news" .3 "new" .3 "last" .25 "show" .2 "display" .2 "see" .2
             "local" .3 "posted" .3 "gave" .3 "show" .3 "personal" .2 "private" .2 
			 "notes" .3 "delete" -1 "add" -1)
  )

;;;=========================================================== GETTING LAST NEWS

(deftask "NEWS: get last news"
  :doc "Task (question) for getting the last news."
  :dialog _get-last-news-conversation
  :indexes ("news" .3 "new" .31 "last" .25 "latest" .25)
  )

(deftask "NEWS: get news"
  :doc "Task (command) for getting the last news."
  :dialog _get-last-news-conversation
  :performative :command
  :indexes ("news" .3 "new" .31 "last" .25 "latest" .25 "show" .2 "display" .2 "see" .2
             "current" .25 "delete" -1 "add" -1)
  )

;;;================================================================ GETTING NEWS

(deftask "NEWS: get news"
  :doc "Task (question) for retrieving news items on a particular subject."
  :dialog _get-news-conversation
  :indexes ("news" .3 "new" .3 "last" .2 
            "fact" .2 "facts" .2 "event" .2 "events" .2   
             "at" .1  "on" .3 "about" .3 "concerning" .3 "from" .3
			"delete" -1 "add" -1)
  )

(deftask "NEWS: get news"
    :doc "Task (command) for retrieving news items on a particular subject."
  :performative :command
  :dialog _get-news-conversation
  :indexes ("show" .3 "see" .2 "on" .3 "about" .3 "concerning" .3 "from" .3
            "delete" -1 "add" -1)
  )

;;;======================================================================== HELP 

(deftask "help"
  :doc "Task (question) printing some general help."
  :dialog _print-help-conversation
  :indexes ("help" .5 "SOS" .6  "lost" .4 
            "what can you do" .4 "what services" .4)
  )

(deftask "help"
  :doc "Task (command) printing some general help."
  :performative :command
  :dialog _print-help-conversation
  :indexes ("help" .5 "SOS" .6  "lost" .4 
            "what can you do" .4 "what services" .4)
  )

;;;================================================================= PROCESS ASK 
;;; Generic task to process ASK messages

;;;(deftask "process ask"
;;;  :doc "Tâche permettant de traiter un message de type ASK."
;;;  :performative :command
;;;  :dialog _process-ask-conversation
;;;  :indexes (":process-ask" 1)
;;;  )
  
;;;================================================================ PROCESS TELL 
;;; Generic task to process TELL messages
;;;   This task displays the content of the message and either keeps it or deletes
;;; it.

;;;(deftask "process tell"
;;;  :doc "Tâche permettant de traiter un message de type TELL."
;;;  :performative :command
;;;  :dialog _process-tell-conversation
;;;  :indexes (":process-tell" 1)
;;;  )
  
;;;======================================================== PUBLISH PRIVATE NOTE
;;; NEWS: making a private note public

(deftask "publish note"
  :doc "Tâche permettant de publier une note privée."
  :performative :command
  :dialog _publish-note-conversation
  :indexes ("publish" .4 "note" .3 "private" .1 "diffuse" .4)
  )
  
;;;=================================================================== SET FONTS 

(deftask "font handler"
  :doc "Tâche permettant de modifier la taille des caractères sur l'écran."
  :performative :command
  :dialog _set-font-size-conversation
  :indexes ("letters" .3 "write" .3)
  )

;;;==================================================================== SUBSCRIBE

(deftask "NEWS: subsribe to NEWS"
  :doc "Task for subscribing to NEWS."
  :performative :command
  :dialog _subscribe-conversation
  :indexes ("subscribe" .5  "receive" .3 "NEWS" .2 "unsubcribe" -1)
  )

;;;========================================================================= TELL

;;;(deftask "tell agent"
;;;  :doc "Tâche (inform) pour demander une information à un autre agent"
;;;  :performative :command
;;;  :dialog _tell-conversation
;;;  :indexes ("informer" .8 "dire à" .7 "information pour" .7 
;;;            "prévenir" .7)
;;;  )

;;;======================================================================= TRACE 

;;;(deftask "set trace"
;;;  :doc "Tâche permettant d'activer ou de supprimer la trace."
;;;  :performative :command
;;;  :dialog _trace-conversation
;;;  :indexes ("transition" .5 "transitions" .5 "trace" .5 "verbose" .5
;;;            "bavard" .5)
;;;  )

;;;================================================================== UNSUBSCRIBE

(deftask "NEWS: unsubsribe to NEWS"
  :doc "Task for unsubscribing to NEWS."
  :performative :command
  :dialog _unsubscribe-conversation
  :indexes ("cancel" .3 "unsubscribe" .3 "delete subscription" .3 "NEWS" .2 
             "subscription" .2 "subscribe" -1)
  )

;;;========================================================== VIEW-SUBSCRIPTIONS

(deftask "NEWS: display all subscriptions"
  :doc "Task for asking PUBLISHER to display all our subscriptions."
  :performative :command
  :dialog _view-subscriptions-conversation
  :indexes ("view" .3 "display" .3 "show" .3 "subscription" .3 "subscriptions" .3 
             "list" .3 "subscribe" .1 "what" .1 "unsubscribe" -1)
  )
  
;;;===================================================================== WHAT-IS 

(deftask "what is"
  :doc "Tâche permettant d'obtenir des infos sur une expression."
  :dialog _what-is-conversation
  :indexes ("what" .2 "is" .25 "what is it?" .7)
  )

;;;====================================================================== WHO-IS 

;;;(deftask "who is"
;;;  :doc "Tâche permettant d'obtenir des infos sur une personne."
;;;  :dialog _get-person-info-conversation
;;;  :indexes ("qui" .5 "est" .2 "c est qui" .7)
;;;  )

;;;============================================================================== 

:EOF