var labelType, useGradients, nativeTextSupport, animate;

(function() {
  var ua = navigator.userAgent,
      iStuff = ua.match(/iPhone/i) || ua.match(/iPad/i),
      typeOfCanvas = typeof HTMLCanvasElement,
      nativeCanvasSupport = (typeOfCanvas == 'object' || typeOfCanvas == 'function'),
      textSupport = nativeCanvasSupport 
        && (typeof document.createElement('canvas').getContext('2d').fillText == 'function');
  //I'm setting this based on the fact that ExCanvas provides text support for IE
  //and that as of today iPhone/iPad current text support is lame
  labelType = (!nativeCanvasSupport || (textSupport && !iStuff))? 'Native' : 'HTML';
  nativeTextSupport = labelType == 'Native';
  useGradients = nativeCanvasSupport;
  animate = !(iStuff || !nativeCanvasSupport);
})();

var Log = {
  elem: false,
  write: function(text){
    if (!this.elem) 
      this.elem = document.getElementById('log');
    this.elem.innerHTML = text;
  }
};

function renderTree(sourceObjs,divName,language){
  
  //Create a new ST instance  
  var st = new $jit.ST({  
    //id of viz container element  
    injectInto: divName,  
    orientation : 'left',
    //set duration for the animation  
    duration: 100,  
    //set animation transition type  
    transition: $jit.Trans.Quart.easeInOut,  
    //set distance between node and its children  
    levelDistance: 60,  
    //enable panning  
    Navigation: {  
      enable:true,  
      panning:true,
      zooming: 10 //zoom speed. higher is more sensible  

    },  
    //set node and edge styles  
    //set overridable=true for styling individual  
    //nodes or edges  
    Node: {  
        height: 30,  
        width: 160,  
        type: 'rectangle',  
        color: '#3366FF',  
        overridable: true,
        'CanvasStyles': {  
          shadowColor: '#ccc',  
          shadowBlur: 10,
        }        
    },        
    Edge: {  
        type: 'arrow',  
        overridable: true  
    },
    onBeforeCompute: function(node){  
        // nothing
    },        
    onAfterCompute: function(){  
        
    },  
      
    //This method is called on DOM label creation.  
    //Use this method to add event handlers and styles to  
    //your node.  
    onCreateLabel: function(label, node){  
        label.id = node.id;              
        label.innerHTML = "<span style='display: inline-block; vertical-align: middle; line-height: normal;'>" + node.name + "</span>";  
        label.onclick = function(){  
					st.onClick(node.id);  
        };  
        //set label styles  
        var style = label.style;  
        style.width = 160 + 'px';  
        style.height = 30 + 'px';      
        style.lineHeight = style.height;
        style.cursor = 'pointer';  
        style.color = '#FFFFFF';  
        style.fontSize = '1.5vh';  
        style.textAlign= 'center';  
        style.verticalAlign = 'bottom';
        //style.paddingTop = '3px';
        
        // set label scaling
        var x = st.canvas.scaleOffsetX,
            y = st.canvas.scaleOffsetY;
    },  
      
    onBeforePlotNode: function(node){
      if(node.data.direct) {
        node.data.$color = "#CC0000";
      }
      // resize labels
      var x = st.canvas.scaleOffsetX,
          y = st.canvas.scaleOffsetY;
          
      $(".node").css("-moz-transform", "scale(" + x +"," +  y + ")");
      $(".node").css("-webkit-transform", "scale(" + x +"," + y + ")");
      $(".node").css("-ms-transform", "scale(" + x +"," +  y + ")");
      $(".node").css("-o-transform", "scale(" + x +"," +  y + ")");
      
      
    },  
    onBeforePlotLine: function(adj){  
      
    }  
  }); 
    
  var filteredObjs = filterConcepts(sourceObjs);
  
  var targetId = getTargetId(filteredObjs);
  
  buildTree(filteredObjs,language);
  
  var jsonTree = createTree(filteredObjs);

  //load json data  
  st.loadJSON(jsonTree);  
  
  //compute node positions and layout  
  st.compute();  
  
  //optional: make a translation of the tree  
  st.geom.translate(new $jit.Complex(-200, 0), "current");  

  //emulate a click on the selected node.  
  st.onClick(targetId);
  
  return st;
}

function filterConcepts(dirtyList) {
	for(var i=dirtyList.length-1;i>=0;i--) {
		var type = dirtyList[i]["marshaller::type"];
		if(type != "CSCWD Concept") {
			dirtyList.splice(i,1);
		}
	}
	return dirtyList;
}

function getTargetId(listObj) {
  for(var i=0;i<listObj.length;i++) {
    if(listObj[i]["marshaller::condition"][0] == "DIRECT") {
      listObj[i].data = {direct:"direct"};
      return listObj[i]["marshaller::uid"];
    }
  }
  return null;
}

function buildTree(listObj,language) {
  var markedForDeletion = [];
  var subsumingConceptName = language == "language=FR" ? "super-type" : "subsuming concept";
  // normalize names
  for(var i=0;i<listObj.length;i++) {
    listObj[i].id = listObj[i]["marshaller::uid"][0];
    listObj[i].name = listObj[i]["marshaller::summary"][0];
    listObj[i].children = [];
  }
  // aggregate parents
  for(var i=0;i<listObj.length;i++) {
    if(listObj[i][subsumingConceptName] != null) {
      var parent = getParent(listObj[i][subsumingConceptName][0],listObj);
      if(parent == null) {
        alert("error processing list: parent not found " + listObj[i][subsumingConceptName][0]);
        return;
      }
      if(!childExists(listObj[i],parent.children)) {
        parent.children.push(listObj[i]);
        markedForDeletion.push(listObj[i]["marshaller::uid"]);
      }
    }
  }
  // aggregate childs
  for(var i=0;i<listObj.length;i++) {
    if(listObj[i]["inverse::" + subsumingConceptName] != null) {
      var childs = listObj[i]["inverse::" + subsumingConceptName];
      for(var j=0;j<childs.length;j++) {
        var child = getElementById(childs[j],listObj);
        if(child == null) {
          alert("Error processing list of objects: child not found " + childs[j]);
          return;
        }
        if(!childExists(child,listObj[i].children)) {
          listObj[i].children.push(child);
          markedForDeletion.push(child["marshaller::uid"]);
        }
      }
    }
  }
  // remove childs marked for deletion
  for(var i=listObj.length-1;i>=0;i--) {
    for(var j=0;j<markedForDeletion.length;j++) {
      if(listObj[i]["marshaller::uid"] == markedForDeletion[j]) {
        listObj.splice(i,1);
        markedForDeletion.splice(j,1);
        break;
      }
    }
  }
}

function childExists(child,children) {
  for(var i=0;i<children.length;i++) {
    if(child["marshaller::uid"] == children[i]["marshaller::uid"]) {
      return true;
    }
  }
  return false;
}

function getElementById(id,list) {
  for(var i=list.length-1;i>=0;i--) {
    if(list[i]["marshaller::uid"] == id) {
      return list[i];
    }
  }
  return null;
}

function getParent(id,obj) {
  if(obj instanceof Array) {
    for(var i=0;i<obj.length;i++) {
      var parent = getParent(id,obj[i]);
      if(parent != null) {
        return parent;
      }
    }
  } else {
    if(id == obj["marshaller::uid"]) {
      return obj;
    }
    if(obj.children != null)
      for(var i=0;i<obj.children.length;i++) {
        var parent = getParent(id,obj.children[i]);
        if(parent != null) {
          return parent;
        }
      }
  }
	return null;
}



/*
 extract the remaining element of the processed list.
 */
function createTree(listObj) {
  if(listObj.length != 1) {
    alert("Error generating tree - #root = " + listObj.length);
    return;
  }
  return listObj[0];
}