﻿;;;===============================================================================
;;;10/11/21
;;;             AGENT ARISTIDE: ONTOLOGY (file ARISTIDE-ontology.lisp)
;;;             Copyright Barthès@UTC, 2009
;;;
;;;=============================================================================== 

#|
ARISTIDE is the French assistant agent of MR.
It can find email addressses, telephone numbers, or use web services (?)

Its ontology is used to process dialogs, answer requests or trigger actions, i.e.
send messages to other agents.
The current ontology contains information about addresses, publications and 
research specialties.
However, ARISTIDE is not a specialist of those subjects and its ontology is rather
shallow. It contains enough vocabulary to ask a specialist (one of the staff agent).

Questions asked to ARISTIDE:
   - what is the telephone number of "Trigano"?
   - phone "dbb"?
   - home phone of "Moulin"?
   - extension of DICIT?
   - send message to "gapenne".
   - ...
   
2009
 0720 adding the description of people/pa combinations in the ARISTIDE package  
2010
 0614 upgrading to v8 agent properties start with ag- 
 0622 a little more clean up
 1121 upgrading to v8.0.8; removing useless concepts
|#

(in-package :ARISTIDE)

(defontology 
  (:title "ARISTIDE ontology")
  (:version "1.0")
  (:language :fr)
  )

;;;-------------------------------------------------------------------------------
;;;                              ETUDIANT
;;;-------------------------------------------------------------------------------		

(defconcept "étudiant"
  (:is-a "personne")
  (:rel "école" (:to "université"))
  )

;;;-------------------------------------------------------------------------------
;;;                               NOUVELLE
;;;-------------------------------------------------------------------------------		

(defconcept "nouvelle"
  (:att "section" (:entry))
  (:att "contenu")
  (:att "date")
  (:att "heure")
  (:att "auteur")
  )

;;;-------------------------------------------------------------------------------
;;;                              PERSONNE
;;;-------------------------------------------------------------------------------		

(defconcept "personne"
  (:att "nom" (:entry))
  (:att "prénom"(:entry))
  (:att "adresse")
  (:att "adresse privée")
  (:att "téléphone")
  (:att "téléphone privé")
  (:att "email")
  (:att "page web")
  (:rel "assistant personnel" (:to "omas agent"))
  )

;;;-------------------------------------------------------------------------------
;;;                              UNIVERSITY
;;;-------------------------------------------------------------------------------		

(defconcept "Université"
  (:att "nom" (:entry))
  )


;;;=============================================================================== 
;;;=============================================================================== 
;;;
;;;                                INDIVIDUALS
;;;
;;;=============================================================================== 
;;;=============================================================================== 

(defindividual "MOSS documentation"  ; class is defined in MOSS
  ("moss title" "global help")
  ("moss documentation" 
   (:fr 
    "Aide globale... En tant qu'assistant, je peux faire un certain nombre ~
     de choses pour vous. La meilleure façon de m'utiliser est de faire des ~
     phrases courtes. Toutefois, je vous rappelle que mon vocabulaire est  ~
     limité."))
  )

;;; ARISTIDE has been defined as an omas-agent when created
;;; ALBERT and ALFREDO have also been defined as OMAS agents but are not in ARISTIDE 
;;; environment

(defindividual "omas agent"
  ("ag name" "albert")
  ("ag key" :albert)
  ;(:language :fr)
  )

(defindividual "omas agent"
  ("ag name" "stevens")
  ("ag key" :stevens)
  ;(:language :en)
  )

(defindividual "omas agent"
  ("ag name" "alfredo")
  ("ag key" :alfredo)
  ;(:language :en)
  )

;;; Note in the following that the value assigned to "assistant personnel" 
;;; is a MOSS query. This is linked to the fact that agent names are  not
;;; entry points. I forgot why they should not be entry points however...

(defindividual "personne"
  ("nom" "Barthès")
  ("prénom" "Jean-Paul")
  ("assistant personnel" ("omas agent" ("ag name" :is "ALBERT")))
  )

(defindividual "personne"
  ("nom" "Gonzales")
  ("prénom" "Omar")
  ("assistant personnel" ("omas agent" ("ag name" :is "alfredo")))
  )

(defindividual "personne"
  ("nom" "Ramos")
  ("prénom" "Milton")
  ("assistant personnel" ("omas agent" ("ag name" :is "aristide")))
  )

;;;=============================================================================== 

:EOF