﻿;;;-*- Mode: Lisp; Package: "XXXXX" -*-
;;;===============================================================================
;;;14/12/03
;;;             SHARED ENGLISH DIALOG (file SHARED-DIALOG-EN.lisp)
;;;             Copyright Barthes@UTC, 2014
;;;
;;;=============================================================================== 

;;; This file contains the dialog shared by English speaking PAs. It is loaded
;;; into their package when creating the agent

#|
2014
 1203 Creation
|#

(in-package :XXXXX)

(format t "~%;*** Loading XXXXX dialogs, package is: ~S" *package*)

;;;================================================================================
;;;
;;;                              Sub-Conversations
;;;
;;;================================================================================

;;; Sub-conversations are defined by a conversation header, that points to the
;;; entry state.
;;; There are two output states:
;;;   SUCCESS: the result is stored into the RESULTS slot of the state-context
;;;   FAILURE: no result could be achieved
;;; In case of bad error there is a throw to an :error label
;;; data are obtained from the HAS-DATA slot of the context object

;;;================================================================================
;;;
;;;                          PRINT HELP CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to help the master by giving information:
;;;   - in general (what tasks are available
;;;      "what can you do for me?" "help." "what can I do?"
;;;   - on a particular topic
;;;      "how do I send a mail?" "help me with the mail?"
;;; specific help corresponding to special help tasks are caught by the task 
;;; task selection mechanism?

(defsubdialog 
  _print-help-conversation 
  (:label "Help conversation")
  (:explanation 
   "Help was asked.")
  (:states _ph-print-global-help)
  )

;;;--------------------------------------------------------- (PH) PRINT-GLOBAL-HELP

(defstate
  _ph-print-global-help
  (:label "Print general help")
  (:entry-state _print-help-conversation )
  (:explanation "No specific subject was included.")
  (:execute
   (let
     ((obj-id (car (send '>-global-help '=get 'is-title-of)))
      (*language* :fr)
      (task-list (access '(task)))
      (conversation (omas::conversation XXXXX::pa_XXXXX))      
      )
     (when obj-id (send obj-id '=get-documentation)
           (send conversation '=display-text *answer*))
     (send conversation '=display-text
           "I can do the following:" :newline t)
     (dolist (task task-list)
       (send task '=get-documentation :lead "   - ")   
       (send conversation '=display-text *answer*)
       )
     (send conversation '=display-text " " :newline t)
     ))
  (:transitions
   (:always :success))
  )
  

;;;================================================================================
;;;
;;;                       SET FONT SIZE CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to modify the size of the dialog font, usually to
;;; increase it or to return to the default value.

(eval-when (compile load eval)
  (proclaim '(special 
              _sfs-entry-state  ; required by defstate
              _sfs-dont-understand
              _sfs-try-again
              _sfs-sorry)))

;;;-------------------------------------------------- (SFS) FONT-SIZE-CONVERSATION

(defsubdialog 
    _set-font-size-conversation
    (:short-name sfs)
  (:label "Font size conversation")
  (:explanation 
   "The data is searched for the presence of words like increase, ~
    bigger, smaller, defaults, reset,...")
  ;(:action _sfs-action)
  )

;;;--------------------------------------------------- (SFS) FONT-SIZE-ENTRY-STATE

(defstate 
    _sfs-entry-state
    (:entry-state _set-font-size-conversation)
  (:label "Font size dialog entry")
  (:explanation "The data is searched for the presence of words like increase, ~
                 bigger, defaults, reset,...")
  (:transitions
   (:patterns ((* "increase" *)
               (* "bigger" *)
               (* "larger" *))
              :exec (omas::assistant-use-larger-font PA_XXXXX) :success)
   (:patterns ((* "decrease" *)
               (* "smaller" *))
              :exec (omas::assistant-reset-font PA_XXXXX) :success)
   (:patterns ((* "use" "default" *)
               (* "reset" *)
               (* "defaults" *))
              :exec (omas::assistant-reset-font PA_XXXXX) :success)
   (:otherwise :target _sfs-dont-understand)
   )
  )

;;; add entry state to dialog header
;(send 'moss::$ESTS '=delete-all _set-font-size-conversation)
;(moss::%link _set-font-size-conversation 'moss::$ESTS _sfs-entry-state)

;;;--------------------------------------------------------- (SFS) DONT-UNDERSTAND

(defstate 
    _sfs-dont-understand
    (:label "Did not understand. Ask master.")
  (:explanation "Assistant is asking master for font-size clarification.")
  (:question "~%- Do you want a bigger font size, a smaller one, or use the default?")
  (:transitions
   (:always :target _sfs-try-again)))

;;;------------------------------------------------------------------- (sfs) SORRY

(defstate 
    _sfs-sorry
    (:label "Font dialog failure.")
  (:explanation "We cannot make things out from the data.")
  (:reset)
  (:answer "~%- Sorry I do not understand what you want me to do with the font?.~%")
  (:transitions
   (:always :failure))
  )

;;;--------------------------------------------------------------- (sfs) TRY-AGAIN

;;; The simplest use is to take text in the HAS-DATA slot of the Q-CONTEXT
;;; and send a message to :SA-ADDRESS

(defstate 
    _sfs-try-again
    (:label "Font size try again")
  (:explanation "we asked master and try to pick up the answer.")
  (:transitions
   (:patterns ((* "bigger" *)
               (* "larger" *))
              :exec (omas::assistant-use-larger-font PA_XXXXX) :success)
   (:patterns ((* "smaller" *))
              :exec (omas::assistant-reset-font PA_XXXXX) :success)
   (:patterns ((* "default" *)
               (* "defaults" *))
              :exec (omas::assistant-reset-font PA_XXXXX) :success)
   (:otherwise :target _sfs-sorry)
   )
  )

;;;===============================================================================
;;;
;;;                       TRACE CONVERSATION
;;;
;;;===============================================================================

;;; this conversation is intended to modify the size of the dialog font, usually to
;;; increase it or to return to the default value.

(eval-when (compile load eval)
  (proclaim '(special 
              _tr-entry-state
              )))

;;;------------------------------------------------------- (TR) TRACE-CONVERSATION

(defsubdialog 
    _trace-conversation
    (:label "Trace conversation")
  (:explanation 
   "Command to toggle the dialog trace."))

;;;-------------------------------------------------------- (TR) TRACE-ENTRY-STATE

(defstate 
    _tr-entry-state
    (:label "Trace dialog entry")
  (:entry-state _trace-conversation)
  (:explanation "We search the input for the trace command.")
  
  (:transitions
   ;; system commands
   (:patterns ((* "show" "transitions" *)
               (* "trace" "transitions" *)
               (* "transitions" "on" *)
               (* "print" "transitions" *))
              :exec (setq moss::*transition-verbose* t
                        moss::*traced-agent* PA_XXXXX) :success)
   (:patterns ((* "no" "trace" *)
               (* "kill" "trace" *)
               (* "trace" "off" *)
               (* "stop" "trace" *)
               (* "no" "verbose" *))
              :exec (setq moss::*verbose* nil
                        moss::*transition-verbose* nil
                        moss::*traced-agent* nil) :success)
   (:patterns ((* "trace" "on" *)
               (* "set" "trace" *)
               (* "start" "trace" *)
               (* "verbose" *))
              :exec (setq moss::*verbose* t
                        moss::*traced-agent* PA_XXXXX) :success)
   (:patterns ((* "kill" "transitions" *)
               (* "hide" "transitions" *)
               (* "transitions" "off" *)
               (* "do" "not" "show" "transitions" *))
              :exec (setq moss::*transition-verbose* nil) :success)
   (:otherwise :failure)
   )
  )

#|
;;; add entry state to dialog header
(send 'moss::$ESTS '=delete-all _trace-conversation)
(moss::%link _trace-conversation 'moss::$ESTS _tr-entry-state)
|#

;;;================================================================================
;;;
;;;                          WHAT-IS CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to obtain and print information about something
;;; corresponding to unformatted data

(defParameter *empty-words* 
  '("the" "a" "an" "of" "from" "to" "those" "these" "that" "s" "?"))

;;;================================================================================

(defsubdialog 
    _what-is-conversation
    (:label "Get info dialog")
  (:explanation "Master is trying to get info about some data.")
  (:states    _what-is-conversation  ; required by defstate
           _wai-entry-state
           _wai-ask-staff
           _wai-get-info
           _wai-process-staff-answer
           _wai-sorry)
  )

;;;---------------------------------------------------------- (WHAT-IS) ENTRY-STATE

(defstate 
    _wai-entry-state
    (:label "What is conversation")
  (:entry-state _what-is-conversation)
  (:explanation "Assistant is trying to get object info from its knowledge base.")
  (:transitions
   ;;keep only relevant data
   (:patterns ((* "what" "is" ?y "?")
               (?y "is" "what" "?" *)
               (?y "what" "is" "it" "?")
               (?y)  ; when entering from print-concept-documentation
               )
              :keep ?y
              :target _wai-get-info)
   (:otherwise :target _wai-sorry)
   )
  )

;;;------------------------------------------------------------ (WHAT-IS) ASK-STAFF
;;; changing strategy to :collect-answers
;;; rather than sending a broadcast right away, maybe we should
;;; 1. look into local KB
;;; 2. look into our agenda (sa-address)
;;; 3. ask BIBLIO
;;; This would avoid strange answers resulting from misinterpretation of the
;;; query.

(defstate 
    _wai-ask-staff
    (:label "Ask staff.")
  (:explanation "Ask the staff for help.")
  (:send-message :to :all :self PA_XXXXX :action :what-is :timeout 2
                 :args `(((:data . ,(read-fact moss::conversation :input))
                          (:language . :en)
                          ))
                 :strategy :collect-answers
                 )
  (:transitions 
   (:always :target _wai-process-staff-answer))
  )

;;;------------------------------------------------------------- (WHAT-IS) GET-INFO

(defstate 
    _wai-get-info
    (:label "Looking into local KB info.")
  (:explanation "Look into own KB to get info.")
  (:answer-analysis)
  )

(defownmethod =answer-analysis _wai-get-info (conversation input)
  "we try first to get info locally and if it fails ask all agents.
Arguments:
   agent: assistant."
  (declare (special *empty-words*))
  (let* (entry-list)
    (catch :return
           ;; clean input of parasitic words
           (setq input (reduce  ; JPB 140820 removing mapcan
                        #'append
                        (mapcar
                            #'(lambda (xx) 
                                (if (member xx *empty-words* :test #'string-equal)
                                    nil
                                  (list xx)))
                          input)))
           (moss::vformat "XXXXX: _wai-get-info /+++ cleaned input:~& ~S" input)
           ;; if nothing left, then go ask staff
           (unless input
             (throw :return `(:transition ,_wai-ask-staff)))
           ;; check whether the resulting words correspond to entry points
           (setq entry-list (moss::find-best-entries input))
           (moss::vformat "XXXXX: _wai-get-info /entry list:~&    ~S" entry-list)
           ;; if nothing left go ask staff
           (unless entry-list 
             (throw :return `(:transition ,_wai-ask-staff)))
           ;; if more than one result, ask for help (not yet)
           ;; get documentation or CV
           ;; print it
           (dolist (item entry-list)
             (send item '=get-documentation :lead "   - ")   
             (send conversation '=display-text *answer*)
             )
           ;; transfer to :success    
           (list :success) 
           )
    ))

;;;------------------------------------------------- (WHAT-IS) PROCESS-STAFF-ANSWER

(defstate 
    _wai-process-staff-answer
    (:label "Process staff answer.")
  (:explanation "Process anwers from the staff agents.")
  (:answer-analysis)
  )

(defownmethod =answer-analysis _wai-process-staff-answer (conversation input)
  "we got an answer from some staff or else the master got impatient.
Arguments:
   agent: assistant."
  (declare (ignore input))
  (let* ((agent (car (HAS-MOSS-AGENT conversation)))
         (answer-list (omas::answer agent))
         (*language* :en)
         )
    ;(format *debug-io* "~&+++ =resume; HAS-TO-DO: ~S" (has-to-do conversation))
    ;; clean up HAS-TO-DO (contains "answer-there")
    ;(setf (has-to-do conversation) nil)
    ;; answer contains either the word "failure" or a list of lists of strings
    ;; (("Jean-Paul Barthes is a professor at UTC ..."
    ;;  "Dominique Barthes-Biesel is a professor at UTC ...")
    ;;  ("Jean-Paul Barthes has NN publications on:  ...")
    ;;  ...)
    ;; 
    (moss::vformat "~&>>> WAIdialog /Answer to the send-request:~% ~S" answer-list)
    ;; failure if :failure is returned or :error on timeout
    (cond
     ((moss::is-answer-failure? answer-list)
      `(:transition ,_wai-sorry))
     ((null answer-list)
      (send conversation '=display-text "I don't know...")
      (list :success))
     (t 
      (moss::vformat " _wai-ask-staff /answer-list: ~& ~S" answer-list)
      (send conversation '=display-text "It is:")
      ;; each item of the answer list is like 
      ;;    ("personne" ("nom" "...")("prénom" "..."))
      (dolist (answer answer-list)
        ;; answer could be "failure" rather than a list of strings
        (unless (and (stringp answer) (string-equal answer "failure"))
          (dolist (item answer)
            (send conversation '=display-text 
                  (concatenate 'string "~%" (car item)))
            (send conversation '=display-text
                  (format nil ": ~{~%     ~{~A~^, ~}~}" (cdr item))))))
      (list :success)))
    ))

;;;---------------------------------------------------------------- (WHAT-IS) SORRY

(defstate 
    _wai-sorry
    (:label "Failure to get info.")
  (:explanation "We could not find anything.")
  (:reset)
  (:text "Sorry, I could not find information for the data.")
  (:transitions 
   (:failure))
  )


;;;================================================================================
;;;
;;;                    WHO IS (GET PERSON INFO) CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to print information about a person.

;;;------------------------------------------------------------- CHECK-ENTRY-POINTS

(defun check-entry-points (input)
  "tries to get entry points by combining words.
Arguments:
   input: list of one or more words
Return:
   nil or a list like 
     (\"barthes\" ($T-PERSON-NAME.OF . $E-PERSON.1) ($T-PERSON-NAME.OF . $E-PERSON.2))"
  (let (result entry entry-list expr-list)
    ;; combine all possible words
    (setq expr-list (make-word-combinations input))
    ;; try all possibilities
    (dolist (expr expr-list)
      ;; entry point ?
      (when (setq entry (moss::%is-entry? expr))
        ;; extract corresponding objects
        (setq result (moss::%%get-objects-from-entry-point entry))
        ;; return to the executing environment
        (moss::vformat "...trying ~S in package ~S as entry...~&   => ~S" 
                       expr *package* result)
        (if result (setq entry-list (append entry-list result)))
        ))
    ;; clean list, eliminating tasks whose performative is different of the required
    ;; one
    (setq entry-list (delete-duplicates entry-list :test #'equal))
    (moss::vformat "check-entry-points /Resulting entry list ~&  => ~S" entry-list)
    (mapcar #'cdr entry-list)))

#|
? (XXXXX::check-entry-points '("telephone" "barthes" ))
;***** ...trying "telephone" in package #<Package "SA-ADDRESS"> as entry...
   => ((MOSS::$ENAM.OF . $E-PHONE))
;***** ...trying "barthes" in package #<Package "SA-ADDRESS"> as entry...
   => (($T-PERSON-NAME.OF . $E-PERSON.2) ($T-PERSON-NAME.OF . $E-PERSON.3)
       ($T-PERSON-NAME.OF . $E-PERSON.4))
;***** check-entry-points /Resulting entry list 
  => ((MOSS::$ENAM.OF . $E-PHONE) ($T-PERSON-NAME.OF . $E-PERSON.2)
      ($T-PERSON-NAME.OF . $E-PERSON.3) ($T-PERSON-NAME.OF . $E-PERSON.4))
($E-PHONE $E-PERSON.2 $E-PERSON.3 $E-PERSON.4)
|#
;;;--------------------------------------------------------- MAKE-WORD-COMBINATIONS

(defun make-word-combinations (word-list)
  ;; when empty list or only one word return
  (unless (cdr word-list) (return-from make-word-combinations word-list))
  ;; otherwise build word combinations
  (let ((input-length (length word-list)) ; e.g. 5
        result)
    (dotimes (nn (1+ input-length)) ; e.g. nn: 0 -> 4
      ;; build cursor to move through the list
      (dotimes (jj nn) ; e.g. jj: 0, then jj: 0->1
        (push
         (apply #'moss::%make-phrase 
                (subseq word-list jj (+ (- input-length nn) jj 1)))
         result)))
    (reverse result)))

#|
? (make-word-combinations '("a" "b" "c" "d"))
("a b c d" "a b c" "b c d" "a b" "b c" "c d" "a" "b" "c" "d")
(moss::%%make-name "a b c" :plain)
A-B-C
|#
;;;================================================================================

(eval-when (compile load eval)
  (proclaim '(special 
              _get-person-info-conversation  ; required by defstate
              _gpi-entry-state
              _gpi-ask-staff
              _gpi-get-info
              _gpi-sorry)))

(defsubdialog 
    _get-person-info-conversation
    (:label "Get person info dialog")
  (:explanation "Master is trying to obtain a telephone number.")
  )

(defstate 
    _gpi-entry-state
    (:label "Get person info conversation")
  (:entry-state _get-person-info-conversation)
  (:explanation "Assistant is trying to get the info from its knowledge base.")
  (:transitions
   ;;keep only relevant data
   (:patterns ((* "who" "is" ?y "?")
               (?y "is" "who" *)
               )
              :keep ?y
              :target _gpi-get-info)
   (:otherwise :target _gpi-sorry)
   )
  )

;;;------------------------------------------------------------- (WHO-IS) ASK-STAFF
;;; changing strategy to :collect-answers
;;; rather than sending a broadcast right away, maybe we should
;;; 1. look into local KB
;;; 2. look into our agenda (ADDRESS)
;;; 3. ask BIBLIO
;;; This would avoid strange answers resulting from misinterpretation of the
;;; query.

(defstate 
    _gpi-ask-staff
    (:label "Ask other agents for help.")
  (:explanation "Ask the staff for help.")
  (:send-message :to :all :self PA_XXXXX :action :who-is :timeout 4
                 :args `(((:data . ,(read-fact moss::conversation :input))
                          (:language . :fr)
                          ))
                 :strategy :collect-answers
                 )
  (:transitions
   (:on-failure :target _gpi-sorry)
   (:otherwise
    :exec (dolist (answer (read-fact moss::conversation :answer))
            ;; answer could be "failure" rather than a list of strings
            (unless (or (and (stringp answer) (string-equal answer "failure"))
                        (null answer))
              (dolist (item answer)
                (send moss::conversation '=display-text item))))
    :success))
  )

;;;-------------------------------------------------------------- (WHO-IS) GET-INFO

(defstate 
    _gpi-get-info
    (:label "Looking into local KB info.")
  (:explanation "Look into own KB to get answer to who-is?.")
  (:answer-analysis)
  )

(defownmethod =answer-analysis _gpi-get-info (conversation input)
  "we got an answer from some staff or else the master got impatient.
Arguments:
   conversation: the conversation object
   input: a copy of the FACTS/input area"
  (let* (entry-list)
    (catch :return
           ;; clean input of parasitic words
           (setq input (reduce  ; JPB 140820 removing mapcan
                        #'append
                        (mapcar
                            #'(lambda (xx) 
                                (if (member xx *empty-words* :test #'string-equal)
                                    nil
                                  (list xx)))
                          input)))
           (moss::vformat "XXXXX: _gpi-get-info /+++ cleaned input:~& ~S" input)
           ;; if nothing left, then ask staff
           (unless input
             (throw :return `(:transition ,_gpi-ask-staff)))
           ; check whether the resulting words correspond to entry points
           (setq entry-list (check-entry-points input))
           (moss::vformat "XXXXX: _gpi-get-info /entry list:~& ~S" entry-list)
           ;; if nothing left go ask staff
           (unless entry-list 
             (throw :return `(:transition ,_gpi-ask-staff)))
           ;; if more than one person ask for help (not yet)
           ;; get documentation or CV
           ;; print it
           (dolist (item entry-list)
             (send item '=get-documentation)
             (send conversation '=display-text *answer*))
           ;; transfer to :success    
           (list :success) 
           )
    ))

;;;----------------------------------------------------------------- (WHO-IS) SORRY

(defstate 
    _gpi-sorry
    (:label "Sorry: no info.")
  (:explanation "We could not find anything.")
  (:reset)
  (:text ("- Sorry, I have not the slightest idea.~%"
          "- I don't know.~%"))
  (:transitions 
   (:failure))
  )

;;;===============================================================================
;;;===============================================================================
;;;
;;;                         MAIN (TOPLEVEL) CONVERSATION
;;;
;;;===============================================================================
;;;===============================================================================

;;; the following conversations are defined thereafter
;;; the entry point of the dialog is _MAIN-CONVERSATION

(eval-when (load compile eval)
  (proclaim '(special 
              _MAIN-CONVERSATION
              _EXPLAIN-CONVERSATION
              _PRINT-ALL-DOCUMENTATION-CONVERSATION
              _PRINT-HELP-CONVERSATION
              _PRINT-OBJECT-DOCUMENTATION-CONVERSATION
              _PROCESS-CONVERSATION
              _SET-FONT-SIZE-CONVERSATION
              _TRACE-CONVERSATION
              ;_PRINT-CONVERSATION
              ;_SHOW-EXAMPLE-CONVERSATION
              ;_WHAT-CONVERSATION 
              )))

;;;================================================================================
;;;
;;;                       MAIN CONVERSATION (CONTROL LOOP)
;;;
;;;================================================================================

;;; create a dialog header to be used for the main conversation

(defindividual 
    MOSS-DIALOG-HEADER
    (HAS-MOSS-LABEL "Main conversation loop")
  (HAS-MOSS-EXPLANATION "This is the main conversation loop.")
  (:var _main-conversation))


;; declare that conversation, replacing default conversation
(setf (omas::dialog-header PA_XXXXX) _main-conversation)

;;; debugging function
(defun cc () (moss::start-conversation _main-conversation))

;;;===== states are declared as global variables
;;; the set of states can be considered as a plan to be executed for conducting the
;;; conversation with the master

;;;--------------------------------------------------------------------------------
(eval-when (compile load eval)
  (proclaim '(special 
              _mc-entry-state
              _mc-get-input
              _mc-more?
              _mc-process
              _mc-sleep)))
;;;--------------------------------------------------------------------------------

;;;--------------------------------------------------------------- (MC) ENTRY-STATE

;(defun ff () (start-conversation _main-conversation :first-time t))
;(defun gg () (start-conversation _main-conversation))

;;; define an entry instance and record it in a global variable.
;;; ****** we should enter a fram on the conversation frame list, to handle
;;; success or failure returns from sub-dialogs. I.e. we should make a transition
;;; on the success or failure state of the main conversation.
;;; entry (_mc-failure _mc-success _mc-entry-state nil) ; no goal 

(defstate _mc-entry-state
    (:entry-state _main-conversation)
  (:label "Dialog start")
  (:explanation 
   "Initial state when the assistant starts a conversation. Send a welcome message ~
    and wait for data. Also entered on a restart following an abort.")
  (:reset-conversation)
  (:text "Warning. This is not a real dialog in the sense that every question ~
            usually does not use the previous context. Thus, referents like ~
            pronouns or ellipsis (referring to part of the answer) won't work.")
  (:question-no-erase
   ("Hello! What can I do for you?"
    "Hi! What would you like to know?"
    "Hi! I will try to do my best to answer your question or ~
     execute your orders, but remember that my IQ is limited."))
  (:transitions 
   (:always :target _mc-process))
  )

;;;----------------------------------------------------------------- (MC) GET-INPUT

(defstate _mc-get-input
    (:label "Get input")
  (:explanation
   "State in which the user inputs his request.")
  (:reset)
  (:question "I'm listening....")
  (:transitions
   (:always :target _mc-process)))

;;;--------------------------------------------------------------------- (MC) MORE?

(defstate _mc-more?
    (:label "More?")
  (:explanation "Asking the user it he wants to do more interaction.")
  (:reset-conversation)
  (:question-no-erase
   ("What else can I do for you?"
    "Is there anything more I can do for you?"
    "Would you like to do something else?"
    "OK.")
   ) 
  (:transitions 
   (:starts-with ("nothing") :target _mc-sleep)
   (:no :target _mc-sleep)
   (:yes :target _mc-get-input)
   (:otherwise :target _mc-process)))

;;;------------------------------------------------------------------- (MC) PROCESS

(defstate _mc-process
    (:label "Process state of main conversation")
  (:explanation
   "Here we call the main process organized as a sub-vconversation.")
  (:transitions
   (:always :sub-dialog _process-conversation
            :success _mc-more?
            :failure _mc-more?)))

;;;--------------------------------------------------------------------- (MC) SLEEP

(defstate _mc-SLEEP
    (:label "Nothing to do")
  (:explanation
   "User said she wanted nothing more.")
  (:reset)
  (:text "OK. I wait until you are ready.")
  (:question-no-erase
   "Wake me up when you want by typing something...")
  (:transitions
   (:always :target _mc-process)))


;;;================================================================================
;;;
;;;                      MAIN CONVERSATION (EXECUTION PART)
;;;
;;;================================================================================

;;; this conversation is intended to process the input from te user by first
;;; determining the performative, then the list of possible tasks

(defsubdialog 
    _process-conversation 
    (:label "Process conversation")
  (:explanation 
   "Processing steps of the main conversation.")
  (:states _mc-eliza
           _mc-find-performative
           _mc-entry-state
           _mc-failure
           _mc-find-performative
           _mc-find-task
           _mc-select-task
           _mc-task-dialog)
  )

;;;================================================================================
;;;                            Service functions
;;;================================================================================

;;;--------------------------------------------------------------------- PRINT-TIME

(defun print-time (conversation &optional (language *language*))
  (multiple-value-bind (second minute hour date month year)
      (decode-universal-time (get-universal-time))
    (send conversation '=display-text
          (case language
            (:en
             (format nil "Current time is ~S:~S:~S and today is the ~S/~S/~S~%"
               hour minute second date month year))
            (:fr
             (format nil 
                 "Il est actuellement ~S:~S:~S et nous sommes aujourd'hui le ~
                      ~S/~S/~S~%"
               hour minute second date month year)))
          )))

;;;================================================================================
;;;                            Conversation states
;;;================================================================================


;;;--------------------------------------------------------- (MC) FIND-PERFORMATIVE

(defstate _mc-find-performative
    (:entry-state _process-conversation)
  (:process-state _main-conversation)
  (:label "Find performative")
  (:explanation 
   "Process what the user said trying to determine the type of performative ~
    among :request :assert :command. Put the result if any into the performative ~
    slot of the conversation object.")
  (:transitions 
   ;; time?
   (:patterns (("what" "time" (?* ?x)))
              :exec (print-time moss::conversation :en) :success)
   ;; now really select performatives
   (:patterns (("what" "is" (?* ?x))
               ("what" "are" (?* ?x))
               ("what" (?* ?x))
               ("when" (?* ?x))
               ("where" (?* ?x))
               ("why" (?* ?x))
               ("who" (?* ?x))
               ("whom" (?* ?x))
               ("whose" (?* ?x))
               ("how" "much" (?* ?x))
               ("how" (?* ?x))
               ("is" (?* ?x))
               ("are" (?* ?x))
               ("were" (?* ?x))
               ("was" (?* ?x))
               ("have" (?* ?x))
               ("had" (?* ?x))
               ("do" (?* ?x))
               ("did" (?* ?x))
               ("can" (?* ?x))
               ("could" (?* ?x))
               ("may" (?* ?x))
               ("might" (?* ?x))
               ((?* ?x) "?")
               )
              :set-performative (list :request)
              :target _mc-find-task)
   (:patterns (((?* ?x) "note" (?* ?y))
               ((?* ?x) "remember" (?* ?y)))
              :set-performative (list :assert)
              :target _mc-find-task)
   (:otherwise 
    :set-performative (list :command)
    :target _mc-find-task))
  )

;;;--------------------------------------------------------------------- (MC) ELIZA

(moss::defstate _mc-ELIZA
    (:label "ELIZA")
  (:explanation
   "Whenever MOSS cannot interpret what the user is saying, ELIZA is called to ~
    do some meaningless conversation to keep the user happy.
    We then record what the user says and resume the analysis.")
  (:eliza)
  ;(:answer "~%What you said is certainly interesting, but I did not understand it.")
  (:transitions
   (:always :target _mc-find-performative))
  )

;;;------------------------------------------------------------------- (MC) FAILURE

(defstate _mc-failure
    (:label "Failure state of main conversation")
  (:explanation
   "Failure state is entered when we return from a sub-dialog with a :failure tag. ~
    We check for more tasks to perform (listed in the task-list slot of the ~
    conversation object). If there are more, we execute the first one. If there ~
    are no more, we go to the more? state.")
  (:answer-analysis)
  )

(defownmethod
    =answer-analysis _mc-failure (conversation input)
  (declare (ignore input))
  (let* ((task-list (HAS-MOSS-TASK-LIST conversation))
         task)
    (if task-list
        (progn
          ;; record next task
          (setq task (pop task-list))
          ;; remove it from task-list
          (send conversation '=replace 'HAS-MOSS-TASK-LIST task-list)
          ;; add it to conversation
          (send conversation '=replace 'HAS-MOSS-TASK (list task))
          ;; transfer to sub-dialog
          `(:transition ,_mc-task-dialog))
      ;; when no more tasks we return :failure and the state variable still contains
      ;; :failure. The crawler will return one more level until hitting a non
      ;; failure return containing a specific state (e.g. the one for the main
      ;; conversation)
      (list :failure)
      )))

;;;----------------------------------------------------------------- (MC) FIND-TASK

(defstate _mc-find-task
    (:label "Find task")
  (:explanation 
   "Process the input to determine the task to be undertaken. Combines the words ~
    from the sentence to see if they are entry points for the index property of ~
    any task. Collect all tasks for which there is an index in the sentence.")
  (:answer-analysis)
  )

(defownmethod
    =answer-analysis _mc-find-task (conversation input)
  "Using input to find tasks If none failure, if one, OK, if more, must ask user to ~
   select one. Checks for an entry point for an index of a task.
Arguments:
   conversation: current conversation
   input: list containing the input as a list of words"
  (let* ((performative (read-fact conversation :PERFORMATIVE))
         task-list)
    (moss::vformat "_mc-find-task /input: ~S, package: ~S" input *package*)
    ;; try to find a task from the input text
    (setq task-list (moss::find-objects 
                     '(task (has-index-pattern (task-index (has-index :is :?))))
                     input :all-objects t)) 
    ;; filter tasks that do not have the right performative
    (setq task-list
          (reduce  ; JPB 140820 removing mapcan
           #'append
           (mapcar
               #'(lambda (xx) 
                   (if (intersection performative (HAS-PERFORMATIVE xx))
                       (list xx)))
             task-list)))
    (moss::vformat "_mc-find-task /task list after performative check: ~S" task-list)
    (cond 
     ;; if empty, ask ELIZA
     ((null task-list)
      `(:transition ,_mc-ELIZA))
     ;; if one, then OK
     ((null (cdr task-list))
      ;; save results, make the task the conversation task
      (setf (HAS-MOSS-TASK conversation) task-list)
      `(:transition ,_mc-task-dialog))
     (t 
      ;; otherwise must select one from the results
      (setf (HAS-MOSS-TASK-LIST conversation) task-list)
      `(:transition ,_mc-select-task)))
    ))

;;;--------------------------------------------------------------- (MC) SELECT-TASK

(defstate 
    _mc-select-task
    (:label "mc select task")
  (:explanation "we have located more than one task. We rank the tasks by computing ~
                 the average weight of the terms in the task index slot. We then ~
                 record the list of tasks into the statte-context task-list slot ~
                 and activate the highest ranking task.")
  (:answer-analysis) 
  )

(defownmethod =answer-analysis _mc-select-task (conversation input)
  "we have located more than one task. We rank the tasks by computing ~
   the MYCIN combination weight of the terms in the task index slot. We then ~
   record the list of tasks into the conversation task-list slot ~
   and activate the highest ranking task."
  ;(declare (ignore input))
  ;; the list of tasks is in the HAS-MOSS-TASK-LIST slot. Tasks are defined in the 
  ;; application package
  (let* ((task-list (HAS-MOSS-TASK-LIST conversation))
         patterns weights result pair-list selected-task level word-weight-list)
    (moss::vformat "_select-task /input: ~S~&  task-list: ~S" input task-list)
    ;; first compute a list of patterns (combinations of words) from the input
    (setq patterns (mapcar #'car (moss::generate-access-patterns input)))
    (moss::vformat "_select-task /input: ~S~&  generated patterns:~&~S" 
                   input patterns)
    ;; then, for each task
    (dolist (task task-list)
      (setq level 0)
      ;; get the weight list
      ;(setq weights (HAS-INDEX-WEIGHTS task))
      (setq weights (moss::%get-INDEX-WEIGHTS task))
      (moss::vformat "_select-task /task: ~S~&  weights: ~S" task weights)
      ;; check the patterns according to the weight list
      (setq word-weight-list (moss::%get-relevant-weights weights patterns))  
      (moss::vformat "_select-task /word-weight-list:~&  ~S" word-weight-list)
      ;; combine the weights
      (dolist (item word-weight-list)
        (setq level (+ level (cadr item) (- (* level (cadr item))))))
      (moss::vformat "_select-task /level: ~S" level)
      ;; push the task and weight onto the result list
      (push (list task level) result)
      )
    (moss::vformat "_select-task /result:~&~S" result)
    ;; order the list
    (setq pair-list (sort result #'> :key #'cadr)) 
    (moss::vformat "_select-task /pair-list:~&  ~S" pair-list)
    ;; keep the first task whatever its score
    (setq selected-task (caar pair-list))
    ;; remove the task that have a weight less than task-threshold (default 0.4)
    (setq pair-list
          (remove nil 
                  (mapcar 
                      #'(lambda (xx) 
                          (if (>= (cadr xx) (omas::task-threshold omas::*omas*)) xx))
                    pair-list)))
    ;; if task-list is empty then return th? first saved task
    ;; this may not be a good policy if the score is too low
    (if (null pair-list)
        (progn
          ;; reset the task-list slot of the conversation object
          (setf (HAS-MOSS-TASK-LIST conversation) nil)
          ;; put the saved task into the task slot
          (setf (HAS-MOSS-TASK conversation) (list selected-task))
          ;; go to task-dialog
          `(:transition ,_mc-task-dialog)
          )
      (progn
        ;; remove the weights
        (setq task-list (mapcar #'car pair-list))
        ;; select the first task of the list
        (setq selected-task (pop task-list))
        (moss::vformat "_select-task /selected task: ~S" selected-task)
        ;; save the popped list in the task-list slot of the conversation object
        (setf (HAS-MOSS-TASK-LIST conversation) task-list)
        (setf (HAS-MOSS-TASK conversation) (list selected-task))
        ;; go to task-dialog
        `(:transition ,_mc-task-dialog)))
    ))  

;;;--------------------------------------------------------------- (MC) TASK-DIALOG

(defstate _mc-task-dialog
    (:label "Task dialog")
  (:explanation 
   "We found a task to execute (in the GOAL slot of the conversation). We activate ~
    the dialog associated with this task.")
  (:answer-analysis)
  )

;;;********** must check for possible problems
(defownmethod
    =answer-analysis _mc-task-dialog (conversation input)
  "We prepare the set up to launch the task dialog.
Arguments:
   conversation: current conversation"
  (declare (ignore input))
  (let* ((task-id (car (send conversation '=get 'HAS-MOSS-TASK))))
    (moss::vformat "_mc-task-dialog /task-id: ~S dialog: ~S" 
                   task-id (HAS-DIALOG task-id))
    ;; we launch the task dialog as a sub-conversation
    ;; the task contains the name of a sub-conversation header, e.g. _get-tel-nb
    `(:sub-dialog ,(car (HAS-DIALOG task-id)) :failure ,_mc-failure)
    ))


;;;================================================================================

;;; here the strategy is to develop a specific dialog for each task. However, we 
;;; could also use a generic dialog consisiting in:
;;;    - getting the list of required parameters
;;;    - for each parameter, trying to extract info from data using predefined
;;;      rules or extraction patterns (Note that this is not much different than
;;;      having a predefined dialog, albeit focused on a single parameter)
;;;    - if some required parameters are missing, ask master about them using 
;;;      predefined questions
;;;    - shipping result to ad hoc staff agent

;;; The generic dialog could be an option offered to the programmer, which requires
;;; a suitable modeling of the tasks.


;;;================================================================================

;;;================================================================================
;;;
;;;                             ELIZA CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to do small talk.

(defParameter *eliza-rules*
  '((("hello")
     "Hi! I am XXXXX."
     "Hello there.")
    (("Hi")
     "Hi! I am XXXXX."
     "Hi.")
    (("hello" "XXXXX" *)
     "Hello Master."
     "Hello Master, pleased to see you up and active.")
    (("Hi" "XXXXX" *)
     "Hi Boss."
     "HI Boss. Nice day isn't it?")
    (("bonjour")
     "Je suis désolé mon français n'est pas encore utilisable. Please use English.")
    (("how" "are" "you" (?* ?x))
     "Very well thank you."
     "Really does my health concern you? I am flattered.")
    (("Bye")
     "I am sorry to see you leaving now..."
     "I was happy to talk with you."
     "Bye...")
    (("Good" "bye" *)
     "Are you really leaving?"
     "I was happy to meet you."
     "Bye...")
    (("So" "long" *)
     "So long."
     "See you soon?")
    (("who"  (?* ?x) "Barthès")
     "Jean-Paul Barthès is a professor of artificial intelligence at the ~
      Université de Technologie de Compiègne in France.")
    (("who"  (?* ?x) "Barthes")
     "Jean-Paul Barthès is a professor of artificial intelligence at the ~
      Université de Technologie de Compiègne in France.
      Note however that Barthès is written with an è.")
    (("forget" "it")
     "OK. No hard feelings.")
    (("thank" "you" (?* ?y))
     "you are welcome.")
    (("thanks")
     "you are welcome."
     "don't mention it.")
    (((?* ?x) "computer" (?* ?y))
     "Do computer worry you?"
     "What do you think about machines?"
     "Why do you mention computers?"
     "What do you think machines have to do with your problem?")
    (((?* ?x) "stupid" (?* ?y))
     "Please, behave!"
     "What you say makes me really sad...")
    (((?* ?x) "does" "not" "work" (?* ?y))
     "Of course ?x works!"
     "What are  you saying that ?x does not work?")
    (("help")
     "you want help about what?")
    (((?* ?x) "sorry" (?* ?y))
     "Please don't apologize"
     "Apologies are not necessary"
     "What feelings do you have when you apologize?")
    (((?* ?x) "I" "remember" (?* ?y))
     "Do you often think of ?y"
     "Does thinking of ?y bring anything else to mind?"
     "What else do you remember?"
     "Why do you recall ?y right now?"
     "What in the present situation reminds you of ?y ?"
     "What is the connection between me and ?y ?")
    (((?* ?x) "do" "you" "remember" (?* ?y))
     "Did you think I would forget ?y ?"
     "Why do you think I should recall ?y by now ?"
     "What about ?y ?"
     "You mentined ?y ?")
    (((?* ?x) "if" (?* ?y))
     "Do you really think its likely that ?y ?"
     "Do you wish that ?y ?"
     "What do you think about ?y ?"
     "Really -- if ?y ?")
    (("what" (?* ?x) "do" "for" "me")
     "So many things..."
     "Currently I do not know."
     "What could please you?")
    (("Certainly")
     "If you say so..."
     "OK.")
    (("do" "you" "like" ?x)
     "I love ?x"
     "Not very much."
     "Quite a bit in fact.")
    
    (((?* ?x) "I" "dreamt" (?* ?y))
     "Really -- ?y"
     "Have you ever fantasized ?y while you were awake ?"
     "Have you dreamt ?y before ?")
    (((?* ?x) "dream" "about" (?* ?y))
     "How do you feel about ?y in reality ?")
    (((?* ?x) "dream" (?* ?y))
     "What does this dream suggest to you ?"
     "Do you dream often ?"
     "What persons appear in your dreams ?"
     "Don't you believe that dream has to do with your problem ?")
    (((?* ?x) "my" "mother" (?* ?y))
     "Who else in your family ?y ?"
     "Tell me more about your family")
    (((?* ?x) "my" "father" (?* ?y))
     "Your father ?"
     "Does he influence you strongly ?"
     "What else comes to mind when you think of your father ?")
    (((?* ?x) "I" "want" (?* ?y))
     "What would it mean if you got ?y ?"
     "Why do you want ?y ?"
     "Suppose you got ?y soon ?")
    (((?* ?x) "I" "am" "glad" (?* ?y))
     "How have I helped you to be ?y ?"
     "What makes you happy just now ?"
     "Can you explain why you are suddenly ?y ?")
    (((?* ?x) "I" "am" "sad" (?* ?y))
     "I am sorry to hear you are depressed"
     "I am sure it's not pleasant to be sad")
    (((?* ?x) "are" "like" (?* ?y))
     "What resemblance do you see between ?x and ?y ?")
    (((?* ?x) "is" "like" (?* ?y))
     "In what way is it that ?x is like ?y ?"
     "What resemblance do you see ?"
     "Could there really be some connection ?"
     (How ?))
    (((?* ?x) "alike" (?* ?y))
     "In what way ?"
     "What similarities are there ?")
    (((?* ?x) "same" (?* ?y))
     "What other connections do you see ?")
    
    (((?* ?x) "I" "was" (?* ?y))
     "Were you really ?"
     "Perhaps I already knew you were ?y"
     "Why do you tell me you were ?y by now ?")
    (((?* ?x) "was" "I" (?* ?y))
     "What if you were ?y ?"
     "Do you think you were ?y ?"
     "What would it meant if you were ?y ?")
    (((?* ?x) "I" "am" (?* ?y))
     "In what way are you ?y ?"
     "Do you want to be ?y ?")
    (((?* ?x) "am" "I" (?* ?y))
     "Do you believe you are ?y ?"
     "Would you want to be ?y ?"
     "You wish I would tell you you are ?y"
     "What would it mean if you were ?y ?")
    (((?* ?x) "am" (?* ?y))
     "Why do you say \"AM?\""
     "I don't understand that")
    (((?* ?x) "are" "you" (?* ?y))
     "Why are you interested in whether I am ?y or not ?"
     "Would you prefer if I weren't ?y ?"
     "Perhaps I am ?y in your fantasies")
    (((?* ?x) "you" "are" (?* ?y))
     "What makes you think I am ?y ?")
    (((?* ?x) "because" (?* ?y))
     "Is that the real reason ?"
     "What other reasons might there be ?"
     "Does that reason seem to explain anything else ?")
    (((?* ?x) "were" "you" (?* ?y))
     "Perhaps I was ?y"
     "What do you think ?"
     "What if I had been ?y ?")
    (((?* ?x) "I" "can't" (?* ?y))
     "Maybe you could ?y now"
     "What if you could ?y ?")
    (((?* ?x) "I" "feel" (?* ?y))
     "Do you often feel ?y ?")
    (((?* ?x) "I" "felt" (?* ?y))
     "What other feelings do you have ?")
    (((?* ?x) "I" (?* ?y) "you" (?* ?z))
     "Perhaps in your fantasy we ?y each other")
    (((?* ?x) "why" "don't" "you" (?* ?y))
     "Should you ?y yourself ?"
     "Do you believe I don't ?y ?"
     "Perhaps I will ?y in good time")
    (((?* ?x) "yes" (?* ?y))
     "You seem quite positive."
     "You are sure ?"
     "I understand.")
    (((?* ?x) "no" (?* ?y))
     "Why not ?"
     "You are being a bit negative"
     "Are you saying \"NO\" just to be negative?")
    
    (((?* ?x) "someone" (?* ?y))
     "Can you be more specific ?")
    (((?* ?x) "everyone" (?* ?y))
     "Surely not everyone."
     "Con you think of anyone in particular ?"
     "Who for example ?"
     "You are thinking of a special person ?")
    (((?* ?x) "always" (?* ?y))
     "Can you think of a specific example ?"
     "When ?"
     "What incident are you thinking of ?"
     "Really -- always ?")
    (((?* ?x) "perhaps" (?* ?y))
     "You do not seem quite certain")
    (((?* ?x) "are" (?* ?y))
     "Did you think there might not be ?y ?"
     "Possibly they are ?y")
    (((?* ?x))
     "I do not understand exactly what you want. You may type: help."
     "I am not sure I understand what you want. You may type: help."
     "Could you rephrase your question please?"
     "Sorry I did not get that. Could you rephrase it? Or type: help."
     "Oops I do not understand. Could you say that differently?"
     "Sorry?"
     )))


#|

;;;========================= preparing mailing lists =============================

;;; Ask MORE INFO

(m-defstate _am-get-info ADDR
            (:label "More info")
            (:explanation "Assistant inferred that master wanted to obtain some sort ~
                 of info and has to find out what kind.")
            (:exec-var data)
            (:immediate-transitions
             (:starts-with (tel fax address email mail) :target _am-ask-address)
             (:starts-with (person paper) :target _am-ask-biblio)
             (:starts-with (mailing) :trim-data (data) :target _am-mailing-kind?)
             (:test (null data) :ask "What do you want exactly?")
             ;; otherwise should branch into the ask-around? state
             (:otherwise :target _am-ask-around?))
            (:transitions
             (:starts-with (tel fax address email mail) :target _am-ask-address)
             (:starts-with (person paper) :target _am-ask-biblio)
             (:starts-with (mailing) :trim-data (input) :target _am-mailing-kind?)))

;;; DETERMINE KIND OF MAILING

;;; missing =add-tp context 'HA?-MAILING-TYPE but in order to do that we should
;;; create and use a MAILING-CONTEXT object
;;; this could be done by creating a context type corresponding to the type of
;;; conversation taking place

(m-defstate _am-mailing-kind? ADDR
            (:label "Mailing kind?")
            (:explanation "We determine that master wants to do a mailing, we need to ~
                 know whether it is through email or postal mail.")
            (:exec-var data)
            (:immediate-transitions
             (:contains (email :email) :trim-data (data) 
                        :trim-data (data) :set-context ('HAS-MAILING-TYPE '(email)) 
                        :target _am-query-info)
             (:contains (postal mail :pmail) 
                        :trim-data (data) :set-context ('HAS-MAILING-TYPE '(pmail)) 
                        :target _am-query-info)
             (:otherwise
              :ask "what kind of mailing do you want: email or postal-mail?"))
            (:transitions
             (:contains (email :email) :set-context ('HAS-MAILING-TYPE '(email))
                        :trim-data (input) :target _am-query-info)
             (:contains (postal mail :pmail) :set-context ('HAS-MAILING-TYPE '(email))
                        :trim-data (input) :target _am-query-info)
             (:otherwise (:set-context ('HAS-MAILING-TYPE '(email)) :target _am-query-info))))

;;; QUERY INFO

(m-defstate _am-query-info ADDR
            (:label "Query info")
            (:explanation "We determined that we want to select mailing on a basis of a query ~
                 So we test if we have enough information to send to the address ~
                 agent for processing the query. Currently, we only check if there ~
                 is anything left in the data area. If so, this will be sued for ~
                 the query. ~
                 A query restricts the range of addresses to the result of a query ~
                 to the A-agent. Please formulate the query (example: \"Barthes\" ~
                 for getting all the persons named Barthès, or \"AFIA\" for getting ~
                 only the members of AFIA).")
            (:exec-var data)
            (:immediate-transitions
             (:contains (:help h ? how? how) :target _am-query-help)
             (:test data :target _am-send-query))
            (:ask "You should now formulate a question to restrict the range of addresses. 
Please formulate the query (example: all persons named \"Barthes\", or 
persons working-for \"AFIA\" for getting only the members of AFIA).")
            (:transitions
             (:contains (:help h ? how? how) :target _am-query-help)
             (:otherwise :save input :target _am-send-query)))

;;; QUERY HELP

(m-defstate _am-query-help ADDR
            (:label "Query help")
            (:explanation "help function to be able to formulate request. Indicates the ~
                 understood vocabulary.")
            (:ask "~&Possible vocabulary:
    all, person, company, firm, organization, name, named, location, living,
    living-at, working, working-for, at, employer, employed-by.
Possible expressions:
    person named \"Barthes\"
    person working-at \"UTC\"
    organization named \"IRIT\"
    person living \"Toulouse\"
    company at \"Compiegne\"
")
            (:transitions
             (:otherwise :save input :target _am-send-query)))

;;; SEND QUERY

;;; the comm process must deliver answer as a list for ust to be able to use the
;;; m-defstate macro

(m-defstate _am-send-query ADDR
            (:label "Send query")
            (:explanation "We collected info for producing a request, now we send it to the ~
                 SA-ADDRESS agent.")
            (:exec-var data)
            (:send-message)
            (:transitions
             (:test (and (typep (car input) 'message) (null (contents (car input))))
                    :target _am-no-result)
             (:test (typep (car input) 'message) :save input 
                    :target _am-process-results)))

;;; NO RESULT

(m-defstate _am-no-result ADDR
            (:label "No result")
            (:explanation "We asked SA-ADDRESS for a list of objects and got none. We should ~
                 rephrase the query.")
            (:ask "~& No answer to your query. Please rephrase it.
Possible vocabulary:
    all, person, company, firm, organization, name, named, location, living,
    living-at, working, working-for, at, employer, employed-by.
Possible expressions:
    person named \"Barthes\"
    person working-at \"UTC\"
    organization named \"IRIT\"
    person living \"Toulouse\"
    company at \"Compiegne\"
")
            (:transitions
             (:otherwise :save input :target _am-send-query)))

;;; PROCESS RESULTS

(m-defstate _am-process-results ADDR
            (:label "Process results")
            (:explanation "We got an answer from the SA-ADDRESS agent from which we must chosse.")
            (:exec-var data)
            (:ask-TSW :title "Select results"
                      :choice-list (cddr (contents (car data)))
                      :why "Select the objects that you want to keep on the mailing list.") 
            ;where do we put the result? in to-do, faking a question
            (:transitions
             (:otherwise :set-context ('HAS-RESULTS (list input)) 
                         :target _am-get-mailing-index)))

;;; GET MAILING INDEX

(m-defstate _am-get-mailing-index ADDR
            (:label "Getting index")
            (:explanation "asking master to give a name for mailing list.")
            (:question "Please give me a name for this mailing list?")
            (:transitions
             (:otherwise :save input :target _am-save-mailing-list)))

|#


(mformat "~%;*** MOSS v~A - XXXXX dialogs loaded ***" moss::*moss-version-number*)

;;;=============================================================================== 

:EOF