﻿;;;-*- Mode: Lisp; Package: "XXXXX" -*-
;;;==============================================================================
;;;14/12/03
;;;             AGENT XXXXX: TASKS (file XXXXX-tasks-EN.lisp)
;;;             Copyright Barthes@UTC, 2014
;;;
;;;==============================================================================

#|
This file contains the tasks shared by the English speaking PAs of the NEWS appli-
cation.

Tasks are currently:
HELP
SET FONTS
WHAT-IS
WHO-IS

2014
 0303 cloned from the STEVENS-TASKS file
 1203 upgraded and cleaned
|#

;(in-package :XXXXX) 

(format t "~&;*** loading English tasks, package is: ~S" *package*)

;;;==============================================================================
;;;   
;;;                               TASK MODEL
;;;
;;;==============================================================================

;;; As soon as the PA finds a task to perform it puts its model into the 
;;; conversation object. 
;;; The CONVERSATION object will contain the info corresponding to the task
;;; arguments, filled during the specific conversation attached to the task.

(defconcept (:en "task index")
  (:doc :en "A TASK INDEX is a property of a task. The index attribute specifies ~
             the term being used, the weight attribute gives its importance for the ~
             task.")
  (:att (:en "index"))
  (:att (:en "weight"))
  )

(definstmethod =summary TASK-INDEX ()
  (list (format nil "~{~A ~}: ~{~S~^ ~}" (has-index)(has-weight))))

(defconcept (:en "task")
  (:is-a moss::$QTSKE) ; used for linking inheritage ?
  (:doc :en "A TASK is a model of task to be performed by the assistant.")
  (:att (:en "task name") (:unique)(:entry))
  (:att (:en "dialog"))
  (:att (:en "where to ask")) ; e.g. :ADDRESS
  (:att (:en "fail question"))
  (:rel (:en "index pattern")(:to "task index"))
  (:att (:en "performative") (:id $PRFT)(:one-of :request :assert :command :answer))
  (:att (:en "message action")) ; e.g. :get-address
  )

(definstmethod =summary TASK ()
  (HAS-TASK-NAME))
	 
;;;==============================================================================
;;;   
;;;                               TASK LIBRARY
;;;
;;;============================================================================== 

;;;======================================================================== HELP 

(deftask "help"
  :doc "Task (question) printing some general help."
  :dialog _print-help-conversation
  :indexes ("help" .5 "SOS" .6  "lost" .4 
            "what can you do" .4 "what services" .4)
  )

(deftask "help"
  :doc "Task (command) printing some general help."
  :performative :command
  :dialog _print-help-conversation
  :indexes ("help" .5 "SOS" .6  "lost" .4 
            "what can you do" .4)
  )

;;;================================================================= PROCESS ASK 
;;; Generic task to process ASK messages

;;;(deftask "process ask"
;;;  :doc "Tâche permettant de traiter un message de type ASK."
;;;  :performative :command
;;;  :dialog _process-ask-conversation
;;;  :indexes (":process-ask" 1)
;;;  )
  
;;;================================================================ PROCESS TELL 
;;; Generic task to process TELL messages
;;;   This task displays the content of the message and either keeps it or deletes
;;; it.

;;;(deftask "process tell"
;;;  :doc "Tâche permettant de traiter un message de type TELL."
;;;  :performative :command
;;;  :dialog _process-tell-conversation
;;;  :indexes (":process-tell" 1)
;;;  )
   
;;;=================================================================== SET FONTS 

(deftask "font handler"
  :doc "Task allowing to modify the font size."
  :performative :command
  :dialog _set-font-size-conversation
  :indexes ("letters" .3 "write" .3)
  )

;;;========================================================================= TELL

;;;(deftask "tell agent"
;;;  :doc "Tâche (inform) pour demander une information à un autre agent"
;;;  :performative :command
;;;  :dialog _tell-conversation
;;;  :indexes ("informer" .8 "dire à" .7 "information pour" .7 
;;;            "prévenir" .7)
;;;  )

;;;======================================================================= TRACE 

;;;(deftask "set trace"
;;;  :doc "Tâche permettant d'activer ou de supprimer la trace."
;;;  :performative :command
;;;  :dialog _trace-conversation
;;;  :indexes ("transition" .5 "transitions" .5 "trace" .5 "verbose" .5
;;;            "bavard" .5)
;;;  )

;;;===================================================================== WHAT-IS 

(deftask "what is"
  :doc "Task allowing to get information about an object if it has documentation."
  :dialog _what-is-conversation
  :indexes ("what" .2 "is" .25 "what is it?" .7)
  )

;;;====================================================================== WHO-IS 

;;;(deftask "who is"
;;;  :doc "Tâche permettant d'obtenir des infos sur une personne."
;;;  :dialog _get-person-info-conversation
;;;  :indexes ("qui" .5 "est" .2 "c est qui" .7)
;;;  )

;;;============================================================================== 

:EOF