﻿F;;;===============================================================================
;;;13/11/21
;;;                               ARIBERTO-DIALOG
;;;                        Copyright Fuckner@UTC, 2013
;;;
;;;=============================================================================== 

;;; This file is meant to be loaded when creating a personal assistant by program.
;;; It should be loaded into the new assistant package.
#|
The current file contains its library of dialogs.

Questions asked to PA (in French):
   - quels sont les projets avec la Chine
   - quels sont les projets ICI en cours 

Sub-Conversations are
   - ELIZA
   - GET-PROJECT

2013
 1121 new version. used shared-dialog from HDSRI project as a source. (MFK)
|#

(format t "~&loading ARIBERTO dialogs, package is: ~S" *package*)

;;; adapting OMAS getstr function to current dialog
(defun getstr (tag alist) (or (omas::getstr tag alist)'("")))

(defparameter *empty-words* 
   '("le" "la" "les" "l" "un" "une" "uns" "unes" "de"
     "du" "d" "quel" "quels" "quelle" "quelles" "est" "sont"))

;;;================================================================================
;;;
;;;                              Sub-Conversations
;;;
;;;================================================================================

;;; Sub-conversations are defined by a conversation header, that points to the
;;; entry state.
;;; There are two output states:
;;;   SUCCESS: the result is stored into the RESULTS slot of the state-context
;;;   FAILURE: no result could be achieved
;;; In case of bad error the dialog is restarted at the top level


;;;===============================================================================
;;;
;;;                       GET PROJECTS CONVERSATION
;;;
;;;===============================================================================

;;; this conversation gets the list of currently active HDS projects. It creates
;;; an HTML file (in the cfsl folder) and returns a string modeling the HTML page

(defsubdialog 
    _get-project-conversation
    (:label "Get projects dialog")
  (:explanation 
   "Master is trying to obtain the list of HDS active projects.")
  (:states _gpj-entry-state
           _gpj-get-all-current-projects
           _gpj-sorry)
  )
;;;-------------------------------------------------------------------------------
;;;               GET PROJECTS CONVERSATION SERVICE FUNCTIONS
;;;-------------------------------------------------------------------------------

;;;--------------------------------------------------------- (GPJ) FORMAT-PROJECTS

(defUn format-projects (project-list web-tag)
  "print a list of projects"
  (if web-tag
      (format-projects-html project-list)
   (format nil "~{~A~^~%~}" (mapcar #'format-single-project project-list))))
   
;;;---------------------------------------------------- (GPJ) PRINT-SINGLE-PROJECT

(defUn format-single-project (project-item)
  "input is something like:
 (\"projet\" (\"pays\" \"Mexique\")
             (\"titre\" \"Doctorat Gonzales\") 
             (\"sigle\") 
             (\"responsable scientifique\" \"Barthès: Jean-Paul\" \"Ramos: Felix\")
             (\"début\" \"2009\") (\"fin\" \"2011\")
             (\"correspondant UTC\" \"Barthès: Jean-Paul\")
             (\"domaine\" \"SMA systèmes multi-agents\" \"TCAO CSCW\"))
   to obtain
     Mexique: Doctorat Gonzales (2009-2011) J.-P. Barthès (SMA ...)
   or something similar for companies."
  (let* ((ll (cdr project-item))
         (title (car (getstr "titre" ll)))
         (acronym (car (getstr "sigle" ll))))
    ;; first cook up title according to existing info	 
    (setq title
          (cond
           ((and title acronym) 
            (format nil "~A (~A)" title acronym))
           ((or title acronym)
            (format nil "~A" (or title acronym)))
           (t "Titre inconnu")))
    ;; then apply formatting	  
    (format nil "~{~A~^, ~}: ~A (~{~A~}-~{~A~}) ~{~A~^, ~} (~{~A~^, ~})" 
           (or (getstr "pays" ll) '("France"))
           title
           (getstr "début" ll)
           (getstr "fin" ll)
           (getstr "participants" ll)
           (or (getstr "domaine" ll) '("non précisé"))
           )))
		              
#|
(setq project-item
      '("projet" ("pays" "Mexique")
        ("titre" "Doctorat Gonzales") 
        ("sigle") 
        ("responsable scientifique" "Barthès: Jean-Paul" "Ramos: Felix")
        ("début" "2009") ("fin" "2011")
        ("correspondant UTC" "Barthès: Jean-Paul")
        ("domaine" "SMA systèmes multi-agents" "TCAO CSCW")))
(format-single-project project-item)
"Mexique: Doctorat Gonzales () (2009-2011)  (SMA systèmes multi-agents, TCAO CSCW)"
|#
;;;--------------------------------------------------- (GPJ) FORMAT-PROJECTS-HTML

(defun format-projects-html (project-list)
  (let ((prop-titles 
         '(("pays" "Pays")("titre" "Titre")("sigle" "Sigle")
           ("début" "Début")
           ("fin" "Fin prévue")
           ("domaine" "Domaine"))))
    (omas::make-html-table prop-titles (mapcar #'cdr project-list) 
                           :header-color "green" :text-color "white")))

;;;--------------------------------------------------- (GPJ) FORMAT-PROJECTS-GRAPH
						   
(defun format-projects-graph (project-list web-tag)
  (declare (ignore web-tag))
  (moss-list-to-json (create-envelope (package-name *package*)
                   (create-graph-message project-list))))

;;;--------------------------------------------------- (GPJ) FORMAT-PROJECTS-CHART				   
				   
(defun format-projects-chart (project-list web-tag)
  (declare (ignore web-tag))
  (moss-list-to-json (create-envelope (package-name *package*)
                         project-list)))				   

;;;-------------------------------------------------------------------------------
;;;                           CONVERSATION STATES
;;;-------------------------------------------------------------------------------

;;;------------------------------------------------------------- (GPJ) ENTRY-STATE

(defstate 
    _get-project-entry-state
    (:entry-state _get-project-conversation)
  (:label "Get project dialog")
  (:explanation "Assistant is sending a free-style message to the PROJECT agent.")
  (:transitions
   ;;look for request for all projects...
   (:patterns ((* "en" "cours" *))
              :target _gpj-get-all-current-projects)
   (:otherwise :target _gpj-sorry))
  )

;;;------------------------------------------------ (GPJ) GET-ALL-CURRENT-PROJECTS

(defstate 
    _gpj-get-all-current-projects
    (:label "Get all current projects")
  (:explanation "Assistant is sending a free-style message to the PROJECT agent.")
  (:send-message 
   :to :PROJECT :action :get-all-current-projects
   :args `(((:data . ,(get-fact :input))
            (:language . :fr)
            (:PATTERN "projet" ("id") ("pays")("titre") ("sigle")
                      ("début") ("fin") ("responsable scientifique" ("personne" ("id") ("nom") ("prénom")))
					  ("membres permanents")
                      ("participants")("domaine") ("durée") )))
   )
  (:execute 
    ;; set facts used by result processing sub-dialog
    (copy-fact :answer :rf-result-to-process)
	(set-fact :rf-categories '("pays" "début" "fin"))
	(set-fact :rf-attributes '("début" "fin" "durée"))
	(set-fact :rf-text-format-function #'format-projects)
	(set-fact :rf-graph-format-function #'format-projects-graph)
	(set-fact :rf-chart-format-function #'format-projects-chart)
  )
  (:transitions
   (:on-failure :target _gpj-sorry)
   (:otherwise
    :sub-dialog _result-format-conversation))
)

;;;------------------------------------------------------------------ (GPDJ) SORRY

(defstate 
    _gpj-sorry
    (:label "Projects number failure.")
  (:explanation "We can't get the info.")
  (:reset)
  (:text "- Désolé, je ne trouve pas les projets demandés.")
  (:transitions (:failure))
  )
   
;;;===============================================================================
;;;
;;;                       DIALOG FOR THE RESULT FORMAT (MFK)
;;;
;;;===============================================================================

;;; this conversation gets the result of a query and asks the user
;;; for the format (textual/tabular, graph or chart.

(defsubdialog 
    _result-format-conversation
    (:label "Result format dialog")
  (:explanation 
   "Master is trying to see his/her query result")
  (:states _rf-entry-state
           _rf-table
		   _rf-graph
		   _rf-chart
		   _rf-category
		   _rf-function
		   _rf-data
		   _rf-summarize
		   _rf-send-response
		   _rf-sorry 
		   _rf-cancel)
)  
  
;;;------------------------------------------------------------- (RF) ENTRY-STATE

(defstate 
  _rf-entry-state
  (:entry-state _result-format-conversation)
  (:label "Result format dialog")
  (:explanation "Assistant is asking user for the format")
  (:execute-preconditions 
    (let (msg)
	  (setq msg (format nil "~%J'ai trouvé ~S résultats." 
	    (length (get-fact :rf-result-to-process))))
      (display-message msg moss::conversation)
	 )
  )
  (:question "Choisissez dans quel format vous souhaitez afficher le resultat: <b>texte/tableau<b>, <b>cartographie</b> ou <b>graphique</b>")
  (:transitions
    (:patterns ((* "texte" *)
                (* "tableau" *))
               :target _rf-table)
    (:patterns ((* "cartographie" *)
                (* "graphe" *))
               :target _rf-graph)
    (:patterns ((* "graphique" *))
               :target _rf-chart)
    (:patterns ((* "canceller" *)
	   	        (* "arreter" *))
               :target _rf-cancel)			   
	(:otherwise 
      :exec (display-message "Pouvez-vous dire cela différemment ?" 
	          moss::conversation)
	  :target _rf-entry-state
	)
  )
)  

(defstate
  _rf-chart
  (:label "Chart visualization")
  (:explanation "Build a chart using the result")
  (:question "quel type de graphique vous souhaitez? <b>camembert<b>, <b>courbes</b> ou <b>colonnes</b>")
  (:transitions
    (:patterns ((* "camembert" *))
               :exec (set-fact :rf-selected-chart-type "pie") 
			   :target _rf-category)
    (:patterns ((* "courbe" *)
                (* "courbes" *))
               :exec(set-fact :rf-selected-chart-type "line") 
			   :target _rf-category)
    (:patterns ((* "colonne" *)
	            (* "colonnes" *))
               :exec (set-fact :rf-selected-chart-type "column") 
			   :target _rf-category)
    (:patterns ((* "canceller" *)
                (* "arreter" *))
               :target _rf-cancel)			   
    (:otherwise 
      :exec (display-message "Pouvez-vous dire cela différemment ?" 
	          moss::conversation)
	  :target _rf-chart)			   
  )
)

(defstate
  _rf-category
  (:label "Chart category")
  (:explanation "Ask user for the chart category")
  (:execute-preconditions
    (let (msg)
	  (setq msg "Choisissez les catégories du graphique. ~% Les catégories disponibles sont: ")
	  (dolist (categ (get-fact :rf-categories))
         (setq msg (concatenate 'string msg "<br><b>" categ "</b>")) 	  
	  )
	  (display-message msg moss::conversation)
	)
  )
  (:question "")
  (:execute 
    (let (selected-categories)
	  (format t "~% :input = ~S" (get-fact :input))
	  (setq selected-categories
	    (get-selected-attributes 
		  (get-fact :input) (get-fact :rf-categories)))
      (set-fact :rf-selected-categories selected-categories))		
  )
  (:transitions
    (:patterns ((* "canceller" *)
                (* "arreter" *))
               :target _rf-cancel)
    (:test 
	  (> (length (get-fact :rf-selected-categories)) 0)
	    :target _rf-function
	)
    (:otherwise
	  :exec (display-message "Pouvez-vous dire cela différemment ?" 
	          moss::conversation)
	  :target _rf-category)
  )
)

(defstate
  _rf-function
  (:label "Function to be applied to the data")
  (:explanation "choose what will e the function to be applied to the serie")
  (:question "quel type de fonction vous souhaitez appliquer sour la donnéé ? <b>nombre<b>, <b>somme</b> ou <b>moyenne</b>")
  (:transitions
    (:patterns ((* "nombre" *))
               :exec 
				 (set-fact :rf-selected-chart-function "count") 
			   :target _rf-data)
    (:patterns ((* "somme" *))
               :exec (set-fact :rf-selected-chart-function "sum") 
			   :target _rf-data)
    (:patterns ((* "moyenne" *))
               :exec (set-fact :rf-selected-chart-function "avg") 
			   :target _rf-data)
    (:patterns ((* "canceller" *)
                (* "arreter" *))
               :target _rf-cancel)
    (:otherwise
	  :exec (display-message "Pouvez-vous dire cela différemment ?" 
	          moss::conversation)
	  :target _rf-function)
  )
)

(defstate
  _rf-data
  (:label "Chart data")
  (:explanation "Ask user for the chart data")
  (:execute-preconditions
    (let (msg)
	  (setq msg "Choisissez l'attribute. ~% Les attribute disponibles sont: ")
	  (dolist (attr (get-fact :rf-attributes))
         (setq msg (concatenate 'string msg "<br><b>" attr "</b>")) 	  
	  )
	  (display-message msg moss::conversation)
	)
  )  
  (:question "")
  (:execute
    (let (selected-attr)
	  (setq selected-attr
	    (get-selected-attributes (get-fact :input) (get-fact :rf-attributes)))
      (set-fact :rf-selected-attribute (car selected-attr)))
  )
  (:transitions
    (:patterns ((* "canceller" *)
                (* "arreter" *))
               :target _rf-cancel)
    (:test
	  (not (null (get-fact :rf-selected-attribute)))
	    :target _rf-summarize
	)
    (:otherwise
	  :exec (display-message "Pouvez-vous dire cela différemment ?" 
	          moss::conversation)
	  :target _rf-data)
  )
)

(defstate
  _rf-summarize
  (:label "Chart summarization")
  (:explanation "Execute grouping function")
  (:execute
    (set-fact :answer
       (create-chart-message
         (get-fact :rf-selected-chart-type)    	   
		 (get-fact :rf-selected-categories)
		 (get-fact :rf-selected-chart-function)
		 (get-fact :rf-selected-attribute)
         (group-moss-list 
	       (get-fact :rf-result-to-process) 
		   (get-fact :rf-selected-categories)
	       (get-fact :rf-selected-attribute) 
		   (get-fact :rf-selected-chart-function))))
  )
  (:transitions
    (:patterns ((* "canceller" *)
                (* "arreter" *))
               :target _rf-cancel)			   
    (:otherwise :target _rf-send-response)
  )
)

(defstate
  _rf-send-response
  (:label "Send response to the end-user")
  (:explanation "Send the data to the end-user in the json format")  
  (:transitions
    (:always 
	  :format-answer (get-fact :rf-chart-format-function) 
	  :success
	)
  )
)

(defstate
  _rf-table
  (:label "Table visualization")
  (:explanation "Show the result in a table")
  (:execute-preconditions
    (copy-fact :rf-result-to-process :answer)
  )
  (:transitions
    (:always 
	  :format-answer (get-fact :rf-text-format-function) 
	  :success
	)
  )
)

(defstate
  _rf-graph
  (:label "Graph visualization")
  (:explanation "Show the results as a graph")
  (:execute-preconditions
    (copy-fact :rf-result-to-process :answer)
  )
  (:transitions
    (:always 
	  :format-answer (get-fact :rf-graph-format-function) 
	  :success)
  )
)

(defstate 
  _rf-sorry
  (:label "Failure to understand the format")
  (:explanation "We haven't understand the format")
  (:reset)
  (:text "Désolé, je ne comprends pas votre demande.")
  (:transitions 
    (:failure))
)

(defstate 
  _rf-cancel
  (:label "Cancel the dialog")
  (:explanation "User is requesting to cancel the dialog")
  (:reset)
  (:text "D'accord. La requete a été annulée")
  (:transitions 
    (:failure))
)
  
;;;===============================================================================
;;;
;;;                      TOP-LEVEL CONVERSATION
;;;
;;;===============================================================================

;;; rewritten Jan 2012 to avoid calling process subdialog, which inserts a 
;;; spurious entry on the salient-features queue. We thus reorganize the state
;;; to be at the same level.

;;; we declare the top-level conversations
;;; the entry point of the dialog is _MAIN-CONVERSATION

(eval-when (:load-toplevel :compile-toplevel :execute)
  (proclaim '(special  _MAIN-CONVERSATION)))

;;; create a dialog header to be used for the main conversation

(defindividual 
    MOSS-DIALOG-HEADER
    (HAS-MOSS-LABEL "Main conversation loop")
  (HAS-MOSS-EXPLANATION "This is the main conversation loop.")
  (:var _main-conversation))

;; declare that particular conversation, replacing default conversation in the
;; agent structure
(setf (omas::dialog-header (omas::agent-being-loaded omas::*omas*))
       _main-conversation)

;;;===============================================================================
;;;                            Service functions
;;;===============================================================================

;;; debugging function
(defun cc () (moss::start-conversation _main-conversation))

;;;-------------------------------------------------------------------- PRINT-TIME

(defun print-time (conversation &optional (language *language*))
  (multiple-value-bind (second minute hour date month year)
      (decode-universal-time (get-universal-time))
    (send conversation '=display-text
          (case language
            (:en
             (format nil "Current time is ~S:~S:~S and today is the ~S/~S/~S~%"
               hour minute second date month year))
            (:fr
             (format nil 
                 "Il est actuellement ~S:~S:~S et nous sommes aujourd'hui le ~
                      ~S/~S/~S~%"
               hour minute second date month year)))
          )))

;;;===============================================================================
;;;                            Conversation states
;;;===============================================================================

;;;===== states are declared as global variables
;;; the set of states can be considered as a plan to be executed for conducting 
;;; the conversation with the master

;;;-------------------------------------------------------------------------------

(eval-when (:compile-toplevel :load-toplevel :execute)
  (proclaim '(special
              _mc-eliza
              _mc-entry-state
              _mc-failure
              _mc-find-performative
              _mc-find-task
              _mc-get-input
              _mc-more?
              _mc-process
              _mc-select-task
              _mc-sleep
              _mc-task-dialog)))

;;;-------------------------------------------------------------------------------

;;;-------------------------------------------------------------- (MC) ENTRY-STATE
;;; ENTRY STATE
;(defun ff () (start-conversation _main-conversation :first-time t))
;(defun gg () (start-conversation _main-conversation))

;;; define an entry instance and record it in a global variable.
;;; ****** we should enter a frame on the conversation frame list, to handle
;;; success or failure returns from sub-dialogs. I.e. we should make a transition
;;; on the success or failure state of the main conversation.
;;; entry (_mc-failure _mc-success _mc-entry-state nil) ; no goal 

;;; Action:
;;;   =EXECUTE
;;;   - get ACTION if any from the dialog header
;;;   - reset conversation: HAS-MOSS-FACTS, HAS-MOSS-TASK, HAS-MOSS-TASK-LIST
;;;   - if ACTION IS non NIL, set conversation HAS-GOAL with a pattern obtaines 
;;;     from sending ACTION a =create-pattern message
;;;     otherwise, reset conversation HAS-GOAL
;;;   - display welcome message
;;;
;;;   <Master enters text>
;;;
;;;   =RESUME
;;;   - if INPUT contains one of the abort commands, throws to :DIALOG-ERROR
;;;   - makes a transition to _MC-PROCESS

(defstate _mc-entry-state
    (:entry-state _main-conversation)
  (:label "Début du dialogue")
  (:explanation 
   "Initial state when the assistant starts a conversation. Send a welcome ~
    message and wait for data. Also entered on a restart following an abort.")
  (:reset-conversation)
  (:text "Attention! Ce dialogue est limité. Mon QI est trés faible...~2%")
  (:question-no-erase
   ("Bonjour ! que puis-je faire pour vous ?"
    "Salut ! Alors on fait quoi aujourd'hui ?" 
    "Bonjour ! Je vais faire de mon mieux pour répondre à vos questions."))
  (:transitions 
   (:always :target _mc-process))
  )

;;;-------------------------------------------------------------------- (MC) ELIZA

(moss::defstate _mc-ELIZA
    (:label "ELIZA")
  (:explanation
   "Whenever MOSS cannot interpret what the user is saying, ELIZA is called to ~
    do some meaningless conversation to keep the user happy. It then records the ~
    master's input and transfers to find performative state.")
  (:eliza)
  (:transitions
   (:always :target _mc-find-performative))
  )

;;;------------------------------------------------------------------ (MC) FAILURE

(defstate _mc-failure
    (:label "Failure state of main conversation")
  (:explanation
   "Failure state is entered when we return from a sub-dialog with a :failure ~
    tag. We check for more tasks to perform (listed in the task-list slot of the ~
    conversation object). If there are more, we execute the first one. If there ~
    are no more, we return with a failure tag.")
  (:answer-analysis)
  )

(defownmethod
    =answer-analysis _mc-failure (conversation input)
  (declare (ignore input))
  (let* ((task-list (HAS-MOSS-TASK-LIST conversation))
         task)
    (if task-list
        (progn
          ;; record next task
          (setq task (pop task-list))
          ;; remove it from task-list
          (send conversation '=replace 'HAS-MOSS-TASK-LIST task-list)
          ;; add it to conversation
          (send conversation '=replace 'HAS-MOSS-TASK (list task))
          ;; transfer to sub-dialog
          `(:transition ,_mc-task-dialog))
      ;; when no more tasks, we go ask for another command or request
      `(:transition ,_mc-more?)
      )))
	  
#|
;; could bee writtena as follows, although, since there is some Lisp code, the
;; use of =answer-analysis is justified
(defstate _mc-failure
    (:label "Failure state of main conversation")
  (:explanation
   "Failure state is entered when we return from a sub-dialog with a :failure ~
    tag. We check for more tasks to perform (listed in the task-list slot of the ~
    conversation object). If there are more, we execute the first one. If there ~
    are no more, we return with a failure tag.")
  (:transitions
    (:test (let ((task-list (HAS-MOSS-TASK-LIST moss::conversation))
                 task)
       (when task-list
          ;; record next task
          (setq task (pop task-list))
          ;; remove it from task-list
          (send moss::conversation '=replace 'HAS-MOSS-TASK-LIST task-list)
          ;; add it to conversation
          (send moss::conversation '=replace 'HAS-MOSS-TASK (list task))
		  ))
		:target _mc-tasl-dialog)
	(:otherwise :target _mc-more)
    )
  )
|#
;;;-------------------------------------------------------- (MC) FIND-PERFORMATIVE

(defstate _mc-find-performative
    (:label "Find performative")
  (:explanation 
   "Processes what the user said trying to determine the type of performative ~
    among :request :assert :command. Put the result if any into the performative ~
    slot of the conversation object.")
  (:transitions 
   ;; time?
   (:patterns (("Quelle" "heure" *))
              :exec (print-time moss::conversation :fr) :target _mc-more?)
   ;; now really select performatives
   (:patterns (("qui" *)
               ("quel" "est" *)
               ("quelle" "est" *)
               ("quels" "sont" *)
               ("quelles" "sont" *)
               ("qu" "est-ce" "que" *)
               ("quoi" *)
               ("quand" *)
               ("où" *)
               ("pourquoi" *)
               ("qui" *)
               ("combien" *)
               ("comment" *)
               ("est-ce" "que" *)
               ("est-il" *)
               ("est" "il" *)    ; interface vocal ?
               ("est-elle" *)
               ("est" "elle" *)
               ("sont-ils" *)
               ("sont" "ils" *)
               ("sont-elles" *)
               ("sont" "elles" *)
               ("a-t-il" *)
               ("a-t-elle" *)
               ("ont-ils" *)
               ("ont-elles" *)
               (* "c" "est" "quoi" *)
               (* "c" "est" "qui" *)
               (* "?")
               )
              :set-performative (list :request)
              :target _mc-find-task)
   (:patterns ((* "notez" *)
               (* "noter" *))
              :set-performative (list :assert)
              :target _mc-find-task)
   (:otherwise 
    :set-performative (list :command)
    :target _mc-find-task))
  )

;;;---------------------------------------------------------------- (MC) FIND-TASK

(defstate _mc-find-task
    (:label "Find task")
  (:explanation 
   "Process the input to determine the task to be undertaken. Combines the words ~
    from the sentence to see if they are entry points for the index property of ~
    any task. Collect all tasks for which there is an index in the sentence.")
  (:answer-analysis)
  )

(defownmethod =answer-analysis _mc-find-task (conversation input)
  "Using input to find tasks. If none, failure; if one, OK; if more, must ask ~
   user to select one. Checks for an entry point for an index of a task.
Arguments:
   conversation: current conversation
   input: list containing the input as a list of words"
  (let* ((performative (read-fact conversation :PERFORMATIVE))
         task-list)
    (moss::vformat "_mc-find-task /performative: ~S, package: ~S, input:~%  ~S" 
                   performative *package* input)
    ;; try to find a task from the input text by using combinations of 3 or less
    ;; successive words
    (setq task-list (moss::find-objects 
                     '(TASK 
                       (HAS-INDEX-PATTERN 
                        (TASK-INDEX (HAS-INDEX :is :?))))
                     input :all-objects t)) 
    (moss::vformat "_mc-find-task /possible tasks:~%  ~S" task-list)
    ;; filter tasks that do not have the proper performative
    (setq task-list
          (mapcan #'(lambda (xx) 
                      (if (intersection performative (HAS-PERFORMATIVE xx))
                          (list xx)))
            task-list))
    (moss::vformat "_mc-find-task /task list after performative check: ~S" task-list)
    (cond 
     ;; if empty, ask ELIZA
     ((null task-list)
      `(:transition ,_mc-ELIZA))
     ;; if one, then OK
     ((null (cdr task-list))
      ;; save results, make the task the conversation task
      (setf (HAS-MOSS-TASK conversation) task-list)
      `(:transition ,_mc-task-dialog))
     (t 
      ;; otherwise must select one from the results
      (setf (HAS-MOSS-TASK-LIST conversation) task-list)
      `(:transition ,_mc-select-task)))
    ))

;;;---------------------------------------------------------------- (MC) GET-INPUT

(defstate _mc-get-input
    (:label "Get input")
  (:explanation
   "State in which the user inputs her/his request.")
  (:reset)
  (:question "Je vous écoute...")
  (:transitions
   (:always :target _mc-process)))

;;;-------------------------------------------------------------------- (MC) MORE?
;;; Action:
;;;   - If answer is "rien", "non" transition to SLEEP
;;;   - if answer is "oui" transition to GET-INPUT
;;;   - anything else sends to PROCESS

(defstate _mc-more?
    (:label "More?")
  (:explanation "Asking the user it he wants to do more interaction.")
  (:reset-conversation)
  (:question-no-erase
   ("Quoi d'autre ?"
   ;;"Que puis-je faire d'autre pour vous ?"
   ;; "~%- Y a-t-il autre chose que je puisse faire pour vous ?"
   ;; "~%- Avez-vous d'autres questions ?"
    "OK. Autre chose ?")
   ) 
  (:transitions 
   (:starts-with ("rien") :target _mc-sleep)
   (:no :target _mc-sleep)
   (:yes :target _mc-get-input)
   (:otherwise :target _mc-process)))

;;;------------------------------------------------------------------ (MC) PROCESS
;;; we keep this useles state as a virtual entry point to processing a task

(defstate _mc-process
    (:label "Process state of main conversation")
  (:explanation
   "A virtual entry point to processing master's input.")
  (:transitions
   (:always :target _mc-find-performative)))

;;;-------------------------------------------------------------- (MC) SELECT-TASK

(defstate 
    _mc-select-task
    (:label "mc select task")
  (:explanation 
   "we have located more than one task. We rank the tasks by computing ~
    the average weight of the terms in the task index slot. We then ~
    record the list of tasks into the statte-context task-list slot ~
    and activate the highest ranking task.")
  (:answer-analysis) 
  )

(defownmethod =answer-analysis _mc-select-task (conversation input)
  "we have located more than one task. We rank the tasks by computing ~
   the MYCIN combination weight of the terms in the task index slot. We then ~
   record the list of tasks into the conversation task-list slot ~
   and activate the highest ranking task."
  ;(declare (ignore input))
  ;; the list of tasks is in the HAS-MOSS-TASK-LIST slot. Tasks are defined in the application
  ;; package
  (let* ((task-list (HAS-MOSS-TASK-LIST conversation))
         patterns weights result pair-list selected-task level word-weight-list)
    (moss::vformat "_select-task /input: ~S~&  task-list: ~S" input task-list)
    ;; first compute a list of patterns (combinations of words) from the input
    (setq patterns (mapcar #'car (moss::generate-access-patterns input)))
    (moss::vformat "_select-task /input: ~S~&  generated patterns:~&~S" 
                   input patterns)
    ;; then, for each task
    (dolist (task task-list)
      (setq level 0)
      ;; get the weight list
      ;(setq weights (HAS-INDEX-WEIGHTS task))
      (setq weights (moss::%get-INDEX-WEIGHTS task))
      (moss::vformat "_select-task /task: ~S~&  weights: ~S" task weights)
      ;; check the patterns according to the weight list
      (setq word-weight-list (moss::%get-relevant-weights weights patterns))  
      (moss::vformat "_select-task /word-weight-list:~&  ~S" word-weight-list)
      ;; combine the weights
      (dolist (item word-weight-list)
        (setq level (+ level (cadr item) (- (* level (cadr item))))))
      (moss::vformat "_select-task /level: ~S" level)
      ;; push the task and weight onto the result list
      (push (list task level) result)
      )
    (moss::vformat "_select-task /result:~&~S" result)
    ;; order the list
    (setq pair-list (sort result #'> :key #'cadr)) 
    (moss::vformat "_select-task /pair-list:~&  ~S" pair-list)
    ;; keep the first task whatever its score
    (setq selected-task (caar pair-list))
    ;; remove the task that have a weight less than task-threshold (default 0.4)
    (setq pair-list
          (remove nil 
                  (mapcar 
                      #'(lambda (xx) 
                          (if (>= (cadr xx) (omas::task-threshold omas::*omas*)) xx))
                    pair-list)))
    ;; if task-list is empty then return the first saved task
    ;; this may not be a good policy if the score is too low
    (if (null pair-list)
        (progn
          ;; reset the task-list slot of the conversation object
          (setf (HAS-MOSS-TASK-LIST conversation) nil)
          ;; put the saved task into the task slot
          (setf (HAS-MOSS-TASK conversation) (list selected-task))
          ;; go to task-dialog
          `(:transition ,_mc-task-dialog)
          )
      (progn
        ;; remove the weights
        (setq task-list (mapcar #'car pair-list))
        ;; select the first task of the list
        (setq selected-task (pop task-list))
        (moss::vformat "_select-task /selected task: ~S" selected-task)
        ;; save the popped list in the task-list slot of the conversation object
        (setf (HAS-MOSS-TASK-LIST conversation) task-list)
        (setf (HAS-MOSS-TASK conversation) (list selected-task))
        ;; go to task-dialog
        `(:transition ,_mc-task-dialog)))
    ))

;;;-------------------------------------------------------------------- (MC) SLEEP

(defstate _mc-SLEEP
    (:label "Nothing to do")
  (:explanation
   "User said she wanted nothing more.")
  (:reset)
  (:text "OK. J'attends que vous soyez prêt.")
  (:question-no-erase 
   "Réveillez moi en tapant ou en disant quelque chose...")
  (:transitions
   (:always :target _mc-process)))

;;;-------------------------------------------------------------- (MC) TASK-DIALOG

(defstate _mc-task-dialog
    (:label "Task dialog")
  (:explanation 
   "We found a task to execute (in the GOAL slot of the conversation). We activate ~
    the dialog associated with this task.")
  ;; we launch the task dialog as a sub-conversation
  ;; the task contains the name of a sub-conversation header, e.g. _get-tel-nb
  (:answer-analysis)
  )

;;;********** must check for possible problems
(defownmethod =answer-analysis _mc-task-dialog (conversation input)
  "We prepare the set up to launch the task dialog.
Arguments:
   conversation: current conversation"
  (declare (ignore input))
  (let* ((task-id (car (send conversation '=get 'HAS-MOSS-TASK))))
    (moss::vformat "_mc-task-dialog /task-id: ~S dialog: ~S" 
                   task-id (HAS-DIALOG task-id))
    ;; we launch the task dialog as a sub-conversation
    ;; the task contains the name of a sub-conversation header, e.g. _get-tel-nb
    `(:sub-dialog ,(car (HAS-DIALOG task-id)) :failure ,_mc-failure 
                  :success ,_mc-more?)
    ))

;;;===============================================================================
;;;
;;;                             ELIZA CONVERSATION
;;;
;;;===============================================================================

;;; this conversation is intended to do small talk.

(defparameter *agent-name* 
  (format nil "~A" (omas::key (omas::agent-being-loaded *omas*))))

(defParameter *eliza-rules*
  (case *language*
    (:fr
     (defcomplexelizarules
         (("hello")
          "Bonjour !"
          "Hello there?")
         (("Bom" "dia")
          "Desculpe ! Não falo Portugues."
          "Bom dia. Por favor fale em frances.")
       (("Salut" *)
        (<> "Salut ! Je m'appelle ~A." *agent-name*)
        "Salut ?")
       (("bonjour")
        (<> "Bonjour ! Je m'appelle ~A." *agent-name*))
       (("bonjour" *)
        "Bonjour, Boss !")
       (("hi")
        "Bonjour ! Je m'appelle SHARED. Mon anglais n'est pas bon.")     
       (("comment" "allez-vous" *)
        "Très bien merci."
        "Vraiment, ma santé vous intéresse ? J'en suis trés flatté.")
       (("qui"  * "Barthès")
        "Jean-Paul Barthès est professeur d'intelligence artificielle à ~
         l'Université de Technologie de Compiègne.")
       (("qui"  * "Barthes")
        "Jean-Paul Barthès est professeur d'intelligence artificielle à ~
         l'Université de Technologie de Compiègne.
         Notez toutefois que Barthès s'écrit avec un è.")
       (("laissez" "tomber")
        "OK. Pas de problème.")
       (("merci" *)
        "Avec plaisir.")
       (("tout" *)
        "Vraiment ?"
        "Quelle exigence !")
       ((* "machine" *)
        "Est-ce que les machines vous dérangent ?"
        "Que pensez-vous des machines ?"
        "Pourquoi mentionnez-vous les machines ?"
        "Les machines ont-elles un rapport avec votre problème ?")
       ((* "stupide" *)
        "S'il vous plaît, restez correct !"
        "Je suis peiné par votre remarque.")
       ((?x "ne" "marche" "pas" *)
        "Bien sûr que ?x marche !"
        "Pourquoi dites-vous que ?x ne marche pas ?")
       (("à" "l" "aide")
        "Vous voulez de l'aide à quel sujet?")
       (("help")
        "Vous voulez de l'aide à quel sujet?")
       ((* "désolé" *)
        "S'il vous plaît, ne vous excusez pas."
        "Des excuses ne sont pas nécessaires"
        "Que ressentez-vous quand vous vous excusez ?")
       ((* "Je" "me" "souviens" ?y)
        "Vous souvenez-vous souvent ?y ?"
        "Vous souvenir ?y vous rappelle-t-il autre chose ?"
        "De quoi d'autre vous souvenez-vous ?"
        "Pourquoi vous souvenez-vous ?y maintenant ?"
        "Qu'est-ce qui dans la situation présente vous fait vous souvenir ?y ?"
        "Quel est le lien entre moi et ?y ?")
       ((* "vous" "rappelez-vous" ?y)
        "Pensez-vous que j'aurais oublié ?y ?"
        "Pourquoi pensez-vous que je devrais me rappeler ?y maintenant ?"
        "Qu'y a-t-il à propos de ?y ?"
        "Vous avez mentionné ?y ?")
       ((* "si" ?y)
        "Pensez-vous que ?y soit probable ?"
        "Souhaitez-vous ?y ?"
        "Que pensez-vous de ?y ?"
        "Vraiment -- si ?y ?")
       (("que" * "faire" "pour" "moi")
        "Tant de choses..."
        "A ce moment précis, je ne vois pas."
        "Qu'est-ce qui vous ferait plaisir ?")
       ((* "J" "ai" "rêvé" ?y)
        "Vraiment -- ?y"
        "Avez-vous déjà rêvé ?y tout eveillé ?"
        "Avez-vous déjà ?y auparavant ?")
       ((* "rêve" *)
        "Qu'est-ce que ce rêve vous suggère ?"
        "Rêvez-vous souvent ?"
        "Quelles personnes apparaissent dans vos rêves ?"
        "Ne pênsez-vous pas que les rêves sont liés à votre problème ?")
       ((* "ma" "mère" ?y)
        "Qui d'autre dans votre famille ?y ?"
        "Parlez moi de votre famille.")
       ((* "mon" "père" *)
        "Votre père ?"
        "A-t-il une grande influence sur vous ?"
        "Qu'est-ce qui vous vient à l'esprit quand vous pensez à votre père ?")
       ((* "Je" "voudrais" ?y)
        "Que signifierait ?y ?"
        "Pourquoi voudriez-vous ?y ?")
       ((* "Je" "suis" "content" ?y)
        "Vous ai-je aidé à être content ?y ?"
        "Qu'est-ce qui vous rend heureux en ce moment ?"
        "Pouvez-vous m'expliquer pourquoi vous êtes tout à coup content ?y ?")
       ((* "Je" "déprime" *)
        "J'en suis fort désolé."
        "je suis sûr que ce n'est pas drïle d'être déprimé.")
       ((?x "est" "comme" ?y)
        "Quelle resemblance voyez-vous entre ?x et ?y ?"
        "En quel sens ?x est comme ?y ?"
        "Pensez-vous qu'il y ait vraiment un lien ?")
       (("ce" "n" "est" "pas" ?x)
        "Cela pourrait-il être ?x ?"
        "Vraiment ?"
        "Est-ce bien sûr ?")
       ((* "pareils" *)
        "Comment ça ?"
        "Quelle resemblance y a-t-il ?")
       ((* "pareil" *)
        "Quel autre lien voyez-vous ?")
       
       ((* "J" "étais" ?y)
        "L'étiez-vous vraiment ?"
        "Je me doutais que vous étiez ?y"
        "Pourquoi me dites-vous que vous étiez ?y maintenant ?")
       ((* "Je" "suis" ?y)
        "En quel sens êtes-vous ?y ?"
        "Souhaitez vous être ?y ?")
       ((* "suis-je" ?y)
        "Croyez-vous être ?y ?"
        "Désirez-vous être ?y ?"
        "Qu'est-ce que cela ferait si vous étiez ?y ?")
       ((* "suis" *)
        "Pourquoi dites-vous \"SUIS\" ?"
        "Je ne comprends pas cela.")
       ((* "êtes-vous" ?y)
        "Pourquoi êtes-vous intéressé par savoir si je suis ?y ou pas ?"
        "Préfèreriez-vous que je ne sois pas ?y ?")
       ((* "vous" "êtes" ?y)
        "Qu'est-ce qui vous fait penser que je suis ?y ?")
       ((* "parce" "que" *)
        "Est-ce la seule raison ?"
        "Quelle autre raison pourrait-il y avoir ?"
        "Est-ce que cette raison peut expliquer autre chose ?")
       ((* "étiez-vous"  ?y)
        "Peut-être étais-je ?y"
        "Qu'en pensez-vous ?"
        "Et si j'avais été ?y ?")
       ((* "je" "ne" "peux" "pas" ?y)
        "Peut-être pouvez-vous ?y maintenant ?"
        "Et si vous pouviez ?y ?")
       ((* "Je" "ressents" ?y)
        "Ressentez-vous ?y souvent ?")
       ((* "J" "ai" "ressenti" *)
        "Qu'avez-vous ressenti d'autre ?")
       ((* "pourquoi" "ne" "pas" ?y)
        "Ne devriez-vous pas ?y vous-même ?"
        "Pensez-vous que je pourrais ?y ?"
        "Peut-petre pourrais-je?y en temps utile")   
       ((* "oui" *)
        "Vous me semblez très positif."
        "Vos êtes sûr ?"
        "Je comprends.")
       ((* "non" *)
        "Pourquoi non ?"
        "Vous me semblez très négatif."
        "Est-ce que vous dites \"NON\" simplement pour être négatif ?")
       ((* "quelqu" "un" *)
        "Pouvez-vous être plus précis ?")
       ((* "tout" "le" "monde" *)
        "Sûrement pas tout le monde."
        "Pensez-vous à quelqu'un en particulier ?"
        "Qui par exemple ?"
        "Vous pensez à quelqu'un de précis ?")
       ((* "toujours" *)
        "Pouvez-vous donner un exemple précis ?"
        "Quand ?"
        "Toujours -- vraiment ?")
       ((* "peut-être" *)
        "Vous n'avez pas l'air tout à fait sûr.")
       ((* "est"  ?y)
        "Vous pensez que cela ne pourrait pas être ?y ?"
        "C'est peut-être ?y")
       ((* "OK" *)
        "Bon, comme vous voudrez.")
       ((* "zouave" *)
        "Bachibouzouk."
        "Moule à gaufre"
        "Restez poli, tonnerre de Brest!")
       ((* "connard" *)
        "Je me plaindrai à mon syndicat..."
        "C'est ça, c'est ça...")
       ((* "imbécile" *)
        "Oh!"
        "Crâne d'oeuf !"
        "Crétin des Alpes !")
       ((* "idiot" *)
        "Pas tant que ça..."
        "Pas plus que la moyenne."
        "Si vous le dites...")
       ((* "crétin" "des" "alpes" *)
        "Il vaut mieux être crétin dans les Alpes que partout ailleurs."
        )
       ((* "crétin" *)
        "Moule à gaufres !"
        "Crème de hareng saur !"
        "C'est pas très gentil..."
        )
       
       
       (((?* ?x))
        "Je ne vois pas ce que vous voulez exactement."
        "Je ne suis pas sûr de comprendre ce que vous voulez."
        "Pouvez-vous reformuler votre question s'il vous plaît ?"
        "Désolé, je n'ai pas compris."
        "Aïe ! Je n'ai pas compris."
        "Pouvez-vous dire cela différemment ?"
        "Pardon ?"
        "Quoi ?"
        "Comment ?"
        "J'ai entendu ?x , mais je ne comprends pas."
        )
       ))
    (otherwise
     '((("hello")
        "Hello! I am TYLER."
        "Hello there?")
       (("Hi" (?* ?x))
        "Hi! Boss."
        "Hi!")
       (("bonjour")
        "Je suis désolé mon français n'est pas encore utilisable. Please use English.")
       (("how" "are" "you" (?* ?x))
        "Very well thank you."
        "Really does my health concern you? I am flattered.")
       (("who"  (?* ?x) "Barthès")
        "Jean-Paul Barthès is a professor of artificial intelligence at the ~
      Université de Technologie de Compiègne in France.")
       (("who"  (?* ?x) "Barthes")
        "Jean-Paul Barthès is a professor of artificial intelligence at the ~
      Université de Technologie de Compiègne in France.
      Note however that Barthès is written with an è.")
       (("forget" "it")
        "OK. No hard feelings.")
       (("thank" "you" (?* ?y))
        "you are welcome."
        "don't mention it.")
       (("thanks")
        "you are welcome."
        "don't mention it.")
       (((?* ?x) "computer" (?* ?y))
        "Do computer worry you?"
        "What do you think about machines?"
        "Why do you mention computers?"
        "What do you think machines have to do with your problem?")
       (((?* ?x) "stupid" (?* ?y))
        "Please, behave!"
        "I am pained by your remark.")
       (((?* ?x) "does" "not" "work" (?* ?y))
        "Of course ?x works!"
        "What are  you saying that ?x does not work?")
       (("help")
        "you want help about what?")
       (((?* ?x) "sorry" (?* ?y))
        "Please don't apologize"
        "Apologies are not necessary"
        "What feelings do you have when you apologize?")
       (((?* ?x) "I" "remember" (?* ?y))
        "Do you often think of ?y"
        "Does thinking of ?y bring anything else to mind?"
        "What else do you remember?"
        "Why do you recall ?y right now?"
        "What in the present situation reminds you of ?y ?"
        "What is the connection between me and ?y ?")
       (((?* ?x) "do" "you" "remember" (?* ?y))
        "Did you think I would forget ?y ?"
        "Why do you think I should recall ?y by now ?"
        "What about ?y ?"
        "You mentined ?y ?")
       (((?* ?x) "if" (?* ?y))
        "Do you really think its likely that ?y ?"
        "Do you wish that ?y ?"
        "What do you think about ?y ?"
        "Really -- if ?y ?")
       (("what" (?* ?x) "do" "for" "me")
        "So many things..."
        "Currently I do not know."
        "What could please you?")
       
       (((?* ?x) "I" "dreamt" (?* ?y))
        "Really -- ?y"
        "Have you ever fantasized ?y while you were awake ?"
        "Have you dreamt ?y before ?")
       (((?* ?x) "dream" "about" (?* ?y))
        "How do you feel about ?y in reality ?")
       (((?* ?x) "dream" (?* ?y))
        "What does this dream suggest to you ?"
        "Do you dream often ?"
        "What persons appear in your dreams ?"
        "Don't you believe that dream has to do with your problem ?")
       (((?* ?x) "my" "mother" (?* ?y))
        "Who else in your family ?y ?"
        "Tell me more about your family")
       (((?* ?x) "my" "father" (?* ?y))
        "Your father ?"
        "Does he influence you strongly ?"
        "What else comes to mind when you think of your father ?")
       (((?* ?x) "I" "want" (?* ?y))
        "What would it mean if you got ?y ?"
        "Why do you want ?y ?"
        "Suppose you got ?y soon ?")
       (((?* ?x) "I" "am" "glad" (?* ?y))
        "How have I helped you to be ?y ?"
        "What makes you happy just now ?"
        "Can you explain why you are suddenly ?y ?")
       (((?* ?x) "I" "am" "sad" (?* ?y))
        "I am sorry to hear you are depressed"
        "I am sure it's not pleasant to be sad")
       (((?* ?x) "are" "like" (?* ?y))
        "What resemblance do you see between ?x and ?y ?")
       (((?* ?x) "is" "like" (?* ?y))
        "In what way is it that ?x is like ?y ?"
        "What resemblance do you see ?"
        "Could there really be some connection ?"
        (How ?))
       (((?* ?x) "alike" (?* ?y))
        "In what way ?"
        "What similarities are there ?")
       (((?* ?x) "same" (?* ?y))
        "What other connections do you see ?")
       
       (((?* ?x) "I" "was" (?* ?y))
        "Were you really ?"
        "Perhaps I already knew you were ?y"
        "Why do you tell me you were ?y by now ?")
       (((?* ?x) "was" "I" (?* ?y))
        "What if you were ?y ?"
        "Do you think you were ?y ?"
        "What would it meant if you were ?y ?")
       (((?* ?x) "I" "am" (?* ?y))
        "In what way are you ?y ?"
        "Do you want to be ?y ?")
       (((?* ?x) "am" "I" (?* ?y))
        "Do you believe you are ?y ?"
        "Would you want to be ?y ?"
        "You wish I would tell you you are ?y"
        "What would it mean if you were ?y ?")
       (((?* ?x) "am" (?* ?y))
        "Why do you say \"AM?\""
        "I don't understand that")
       (((?* ?x) "are" "you" (?* ?y))
        "Why are you interested in whether I am ?y or not ?"
        "Would you prefer if I weren't ?y ?"
        "Perhaps I am ?y in your fantasies")
       (((?* ?x) "you" "are" (?* ?y))
        "What makes you think I am ?y ?")
       (((?* ?x) "because" (?* ?y))
        "Is that the real reason ?"
        "What other reasons might there be ?"
        "Does that reason seem to explain anything else ?")
       (((?* ?x) "were" "you" (?* ?y))
        "Perhaps I was ?y"
        "What do you think ?"
        "What if I had been ?y ?")
       (((?* ?x) "I" "can't" (?* ?y))
        "Maybe you could ?y now"
        "What if you could ?y ?")
       (((?* ?x) "I" "feel" (?* ?y))
        "Do you often feel ?y ?")
       (((?* ?x) "I" "felt" (?* ?y))
        "What other feelings do you have ?")
       (((?* ?x) "I" (?* ?y) "you" (?* ?z))
        "Perhaps in your fantasy we ?y each other")
       (((?* ?x) "why" "don't" "you" (?* ?y))
        "Should you ?y yourself ?"
        "Do you believe I don't ?y ?"
        "Perhaps I will ?y in good time")
       (((?* ?x) "yes" (?* ?y))
        "You seem quite positive."
        "You are sure ?"
        "I understand.")
       (((?* ?x) "no" (?* ?y))
        "Why not ?"
        "You are being a bit negative"
        "Are you saying \"NO\" just to be negative?")
       
       (((?* ?x) "someone" (?* ?y))
        "Can you be more specific ?")
       (((?* ?x) "everyone" (?* ?y))
        "Surely not everyone."
        "Con you think of anyone in particular ?"
        "Who for example ?"
        "You are thinking of a special person ?")
       (((?* ?x) "always" (?* ?y))
        "Can you think of a specific example ?"
        "When ?"
        "What incident are you thinking of ?"
        "Really -- always ?")
       (((?* ?x) "perhaps" (?* ?y))
        "You do not seem quite certain")
       (((?* ?x) "are" (?* ?y))
        "Did you think there might not be ?y ?"
        "Possibly they are ?y")
       (((?* ?x))
        "I do not understand exactly what you want. You may type: help."
        "I am not sure I understand what you want. You may type: help."
        "Could you rephrase your question please?"
        "Sorry I did not get that. Could you rephrase it? Or type: help."
        "Oops I do not understand. Could you say that differently?"
        "Sorry?"
        )))
    ))

(format t "~%;*** MOSS v~A - SHARED dialogs loaded ***" moss::*moss-version-number*)

;;;=============================================================================== 

:EOF