﻿;;;-*- Mode: Lisp; Package: "PROJECT" -*-
;;;===============================================================================
;;;13/11/21
;;;                               AGENT PROJECT
;;;                        Copyright Fuckner@UTC, 2010-2013
;;;===============================================================================

;;; Skills for answering the PA
;;;   GET-PROJECT (PA)
#|
2013
 1121 Creation based on project.lisp from HDSRI project.
|#

(defpackage :PROJECT (:use :cl :moss :omas))
(in-package :PROJECT)
 
(omas::defagent :PROJECT :language :fr :persistency t)

;;;===============================================================================
;;; 
;;;                                   SKILLS 
;;;
;;;===============================================================================

;;;========================================================================= Skill
;;;                       GET ALL CURRENT PROJECTS (Pa)
;;;===============================================================================

(defskill :get-all-current-projects :project
  :static-fcn static-get-all-current-projects
  )

(defun static-get-all-current-projects (agent message args)
  "we receive a request for returning all (active) projects, i.e. projects ending ~
   after the current year.
arguments:
   agent: :project
   environment: incoming message
   args: e.g. ((:data \"of\" \"barthes\")(:pattern ...))
return:
   :failure or list of contacts following pattern."
  (declare (ignore message)(special moss::*query-allow-sub-classes*))
  (let ((pattern (cdr (assoc :pattern args)))
        (input-text (cdr (assoc :data args)))
        results reason)	
    
    ;; try first to get projects from entry points in the input text
    ;; when data does not contain entry-points,
    ;; then reason is :no-path. data is not discriminating
    (multiple-value-setq (results reason)
      (locate-objects input-text "projet" 4))
    
    ;; eventually remove theses
    (when results
      (setq results 
            (remove-if #'(lambda (xx) (moss::%type? xx '$E-thesis)) results)))
    
    ;; when we could not get results (no entry point) then get all projects
    (when (eql reason :no-path)
      ;; howwever remove the theses
      (let ((moss::*query-allow-sub-classes* nil))
        (setq results (access '("projet")))))
    
    ;; remove projects that were not accepted
    (setq results 
          (remove-if #'(lambda (xx) (equal (car (send xx '=get "état")) "refusé"))
                     results))
    
    ;; current means that the end-date is greater or equal than current year
    (setq results 
          (keep-after (1- (get-current-year)) "fin" results))
    
    ;; record results into the last-selection slot
    (save-selection agent results)
    (when results 
      (setq results 
            (if pattern 
                (fill-pattern (delete-duplicates results) pattern)
              (broadcast results '=get-project)))
      )
    
    (static-exit agent (or results "*failure*"))))
	
;;;-------------------------------------------------------------------- KEEP-AFTER

(defun keep-after (ref-year prop-ref object-list)
  "remove from object-list all objects with end year after ref-year.
Arguments:
   ref-year: a year as an integer, e.g. 2010
   prop-ref: a prop name e.g. \"date de fin\"
   object-list: a list of object ids
Return:
   the filtered list."
  (let (year result)
    (dolist (obj-id object-list)
	  (setq year (car (send obj-id '=get prop-ref)))
	  (if (stringp year) (setq year (read-from-string year)))
	  (if (and (numberp year) (> year ref-year))
	     (push obj-id result))
	   ) 
   (reverse result)))	

#|
(static-get-all-current-projects SA_PROJECT :message '(((:DATA "projets" "en" "cours" "?") (:LANGUAGE . :FR)
  (:PATTERN "projet" ("pays") ("titre") ("sigle") ("début") ("fin")
   ("participants") ("domaine")))))
|#
	 

:EOF