﻿;;;-*- Mode: Lisp; Package: "WEB" -*-
;;;===============================================================================
;;;13/11/21
;;;                           AGENT WEB
;;;                    Copyright Fuckner@UTC, 2013
;;; 
;;;===============================================================================
;;; this agent is a go between between the web server and PA. It receives 
;;; messages and forwards the message to the PA. It then transfers the
;;; answer to the server thread.
#|
2013
 1121 used HSDLAB web.lisp as a base to create the example application (mfk)
|#

(defpackage :WEB (:use :cl :moss :omas :excl :net.aserve))
(in-package :WEB)

;===============================================================================
;;;                           Creating web agent
;;;===============================================================================

(omas::defagent :WEB)

;;;===============================================================================
;;;                                      Globals
;;;===============================================================================

(defparameter *web-server-port* 80 "server port for serving HDSRI forms.")

;;;===============================================================================
;;;                                    skills 
;;;===============================================================================

;;;========================================================================= skill
;;;                             :PROCESS-WEB-TEXT
;;;===============================================================================

(defskill :PROCESS-WEB-TEXT :WEB
   :static-fcn static-process-web-text
   :dynamic-fcn dynamic-process-web-text)
   
(defun static-process-web-text (agent message args)
  "function processes the data received by the web server to send the text to the
   target PA.
Arguments:
   agent: this agent
   message: incoming message
   args: ((:data ...)(:gate ...)(:to-agent ...)
     - data contains the text to be processed
     - gate contains the index of a synchronization variable. The function waiting
       for the answer is waiting on the gate associated with this index.
     - to-agent is the agent key of the destination PA
Return:
   :done"
  (declare (ignore message))
  (let ((pa (cdr (assoc :to-agent args)))
        (text (cdr (assoc :data args)))
        (gate (cdr (assoc :gate args)))
        (timeout (cdr (assoc :timeout args))))
    (env-set agent gate :gate)
    (send-subtask agent :to pa :action :process-web-text 
                  :args `(((:data . ,text)
                           (:timeout . ,timeout)))
                  :timeout timeout  ; to kill process if it waits too long
                  )
    (static-exit agent :done)))

(defun dynamic-process-web-text (agent message answer)
  (declare (ignore message))
  ;; recover gate index
  (let ((index (env-get agent :gate)))
    ;; save answer
    (omas::web-push-answer index answer)
    (format t "~%;WEB/ dynamic-process-web-text /answer-list: ~S" 
      omas.web::*web-answer-list*)
    ;; open gate
    (omas::web-open-gate index)
    (dynamic-exit agent :done)))

;;;======================================================================== skill
;;;                              :START-SERVER
;;;==============================================================================
;;; this enables the server and loads the entities for answering questions and
;;; filling the forms through web pages
;;; use CONNECT instead, since CONNECT published the needed pages

(defskill :START-SERVER :WEB
   :static-fcn static-start-server)
   
(defun static-start-server (agent message)
  "start the HTTP server, and set up receiving pages for handling the requests"
  (declare (ignore message)
           (special *web-server-port*))
  (let ()
    ;; first start server
    (omas::start-web-server agent *web-server-port*)
    ;; publish a few pages
    (publish-web-query-request agent)
    (net.aserve::publish-directory 
     :prefix "/html" 
     ;       :destination "OMAS-MOSS 10.0.1 (ACL 9.0)\\OMAS\\applications\\UTC-GRAPH_CHART_EXAMPLE\\WEBRES")
     :destination
     "OMAS-MOSS 10.0.2\\OMAS\\sample-applications\\UTC-GRAPH_CHART_EXAMPLE\\WEBRES")
    (static-exit agent :done)))

#|
(static-start-server sa_web nil)
|#
;;;==============================================================================
;;;                                WEB PAGES
;;;==============================================================================

(defun publish-web-query-request (agent) 
  (declare (ignore agent)) 
  (publish
   :path "/webqueryrequest" 
   :content-type "text/html" 
   :function 
   #'(lambda (req ent)
       (let ((user-input (cdr (assoc "user-input" 
                                     (request-query req) :test #'equal))) 
             data text) 
         (with-http-response (req ent)
           (with-http-body (req ent) 
             (setq data
                   (omas::get-web-answer :WEB user-input :to-agent :ARIBERTO)) 
             (setq text (cdr (assoc :text data))) 
             (format (request-reply-stream req) text)))))))			   


;;;==============================================================================
;;;                                    goals
;;;==============================================================================

;;; Goal to start the server a few seconds after the agent has been created

(defgoal :start-server :web 
  :activation-delay 3
  :script start-server-script)

(defUn start-server-script (agent)
  "send a request message to start the web server"
  (declare (ignore agent))
  (list (make-instance 'omas::message :type :request :to :WEB 
                       :action :start-server)))

:EOF